'''
A dialog to replace QMessageBox to display some error messages.
The default QMessageBox is very difficult to control its size.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5.QtGui import QFont
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QDialogButtonBox, QStyle
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QSize


class FailDialog(QtWidgets.QDialog):
    create_fail_summary=pyqtSignal()

    def __init__(self,main_text='',info_text='',detailed_text='',parent=None):
        '''
        Args:
            main_text (str): main text set by setText().
            info_text (str): text set by setInformativeText().
            detailed_text (str): text set by setDetailedText().
            parent (QWidget): parent widget.
        '''

        super(self.__class__,self).__init__(parent=parent)

        self.main_text=main_text
        self.info_text=info_text
        self.detailed_text=detailed_text

        self.setWindowTitle('Error')
        self.resize(400,200)
        self.grid=QtWidgets.QGridLayout(self)

        icon=self.style().standardIcon(QStyle.SP_MessageBoxWarning)
        icon_label=QtWidgets.QLabel()
        icon_label.setPixmap(icon.pixmap(QSize(32,32)))
        self.grid.addWidget(icon_label,0,0)

        self.main_text_label=QtWidgets.QLabel(main_text)
        label_font=QFont('Serif',12,QFont.Bold)
        self.main_text_label.setFont(label_font)
        self.grid.addWidget(self.main_text_label,0,1)

        self.info_text_label=QtWidgets.QLabel(info_text)
        self.info_text_label.setTextFormat(Qt.RichText)
        self.info_text_label.setWordWrap(True)
        self.grid.addWidget(self.info_text_label,1,1,1,2)

        self.buttons=QDialogButtonBox(self)

        self.ok_button=QtWidgets.QPushButton('Ok')
        self.ok_button.setDefault(True)
        self.ok_button.setAutoDefault(True)

        self.detail_button=QtWidgets.QPushButton('Show Details')
        self.detail_button.setDefault(False)
        self.detail_button.setAutoDefault(False)
        self.detail_button.clicked.connect(self.detailButtonClicked)

        self.create_folder_button=QtWidgets.QPushButton('Show Failed Docs')
        self.create_folder_button.setDefault(False)
        self.create_folder_button.setAutoDefault(False)
        self.create_folder_button.clicked.connect(self.createFailList)

        self.buttons.addButton(self.ok_button,QDialogButtonBox.AcceptRole)
        self.buttons.addButton(self.detail_button,QDialogButtonBox.ActionRole)
        self.buttons.addButton(self.create_folder_button,
                QDialogButtonBox.ActionRole)

        self.grid.addWidget(self.buttons,2,2)

        self.text_edit=QtWidgets.QTextEdit(self)
        self.text_edit.setText(detailed_text)
        self.text_edit.setVisible(False)

        self.grid.addWidget(self.text_edit,3,0,1,3)

        self.buttons.accepted.connect(self.accept)

        #self.exec_()

    def setText(self,text):
        self.main_text=text
        self.main_text_label.setText(text)
        return

    def setInformativeText(self,text):
        self.info_text=text
        self.info_text_label.setText(text)
        return

    def setDetailedText(self,text):
        self.detailed_text=text
        self.text_edit.setText(text)
        return


    def detailButtonClicked(self):
        if self.detailed_text=='':
            return

        if self.text_edit.isVisible():
            self.detail_button.setText('Show Details')
            self.text_edit.setVisible(False)
        else:
            self.detail_button.setText('Hide Details')
            self.text_edit.setVisible(True)

        return

    def createFailList(self):
        self.create_fail_summary.emit()
        return



'''
Widget for computing and displaying duplicate checking.

MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QSize
from PyQt5.QtGui import QBrush, QColor, QIcon, QCursor
from PyQt5.QtWidgets import QDialogButtonBox
from ..tools import fuzzyMatch, dfsCC
from .threadrun_dialog import Master


LOGGER=logging.getLogger(__name__)



class CheckDuplicateFrame(QtWidgets.QScrollArea):

    del_doc_from_folder_signal=pyqtSignal(list, # docids
            str, #foldername
            str, #folderid
            bool #reload_talbe
            )
    del_doc_from_lib_signal=pyqtSignal(list, # docids
            bool #reload_table
            )

    def __init__(self,settings,parent=None):
        '''
        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py
        '''
        super(CheckDuplicateFrame,self).__init__(parent=parent)

        self.settings=settings
        self.parent=parent
        self.min_score=self.settings.value('duplicate_min_score',type=int)

        frame=QtWidgets.QWidget()
        self.setWidgetResizable(True)
        self.setWidget(frame)
        va=QtWidgets.QVBoxLayout(self)

        #----------------Create clear frame----------------
        va.addWidget(self.createClearDuplicateFrame())

        #----------------Create treewidget----------------
        self.tree=QtWidgets.QTreeWidget(self)
        self.tree.setColumnCount(7)
        self.tree.setColumnHidden(6,True)

        self.tree.setHeaderLabels(['Group', 'Authors', 'Title',
            'Publication', 'Year', 'Similarity','id'])
        self.tree.setColumnWidth(0, 55)
        self.tree.setColumnWidth(1, 250)
        self.tree.setColumnWidth(2, 300)
        self.tree.setColumnWidth(3, 150)
        self.tree.setColumnWidth(4, 50)
        self.tree.setColumnWidth(5, 20)
        self.tree.setColumnWidth(6, 0)
        self.tree.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.tree.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.tree.header().setStretchLastSection(False)
        self.tree.header().setSectionResizeMode(
                QtWidgets.QHeaderView.Interactive)
        self.tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.tree.setDragDropMode(QtWidgets.QAbstractItemView.NoDragDrop)
        self.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.docTreeMenu)

        self.noDupLabel=QtWidgets.QLabel('No duplicates found.')
        va.addWidget(self.noDupLabel)
        va.addWidget(self.tree)

        frame.setLayout(va)


    def createClearDuplicateFrame(self):

        frame=QtWidgets.QFrame()
        frame.setStyleSheet('background: rgb(235,225,190)')
        ha=QtWidgets.QHBoxLayout()

        # del button
        self.del_duplicate_button=QtWidgets.QToolButton(self)
        self.del_duplicate_button.setText('Delete Selected')
        self.del_duplicate_button.clicked.connect(self.delDocs)

        # clear button
        self.clear_duplicate_button=QtWidgets.QToolButton(self)
        self.clear_duplicate_button.setText('Exit')

        self.clear_duplicate_label=QtWidgets.QLabel('Clear current filtering')
        ha.addWidget(self.clear_duplicate_label)
        tip_label=QtWidgets.QLabel()
        tip_icon=QIcon.fromTheme('help-about').pixmap(QSize(16,16))
        tip_label.setPixmap(tip_icon)
        tip_label.setToolTip('''Change "Mininimum Similary Score" in "Preferences" to change the filtering of matching results.''')
        ha.addWidget(tip_label)
        ha.addWidget(self.del_duplicate_button)
        ha.addWidget(self.clear_duplicate_button)

        frame.setLayout(ha)

        return frame


    def checkDuplicates(self, meta_dict, current_folder, docids1, docid2=None):
        """Start the duplicate checking among documents

        Args:
            meta_dict (dict): meta data of all documents. keys: docid,
                              values: DocMeta dict.
            current_folder (tuple): (foldername_in_str, folder_id_in_str).
            docids1 (list): ids of docs in group 1.

        Kwargs:
            docid2 (int, list or None): if int or list, ids of docs in group 2,
                then duplicate checking is done among all pairs across group 1,
                2. If None, checking is done among all pairs in group 1.
        """

        self.tree.clear()
        self.setVisible(True)

        self.meta_dict=meta_dict
        self.current_folder=current_folder # (name, id)
        self.docids1=docids1
        self.docids1.sort()
        self.docid2=docid2

        self.scores_dict={}

        self.master1=Master(self.prepareJoblist, [(0,self.docids1,self.docid2)],
                1, self.parent.progressbar,
                'busy', self.parent.status_bar, 'Preparing job list...')
        self.master1.all_done_signal.connect(self.jobListReady)
        self.master1.run()

        return


    def prepareJoblist(self, jobid, docids1, docid2):
        '''Prepare fuzzy matching jobs

        Args:
            jobid (int): dummy job id.
            docids1 (list): ids of docs in group 1.
            docid2 (int, list or None): ids of docs in group 2,

        Returns:
            rec (int): 0 for success.
            jobid (int): dummy job id.
            job_list (list): list of tuples, each providing the args for a
                             fuzzyMatch() call.
        '''

        n=len(docids1)
        job_list=[]

        #----------------Check among docds----------------
        if docid2 is None:
            for ii in range(n):
                docii=docids1[ii]
                for jj in range(n):
                    docjj=docids1[jj]
                    if ii>=jj:
                        self.scores_dict[(docii, docjj)]=0
                    else:
                        job_list.append((ii, docii, docjj,
                            self.meta_dict[docii], self.meta_dict[docjj]))
        #-----------------nxm compare-----------------
        else:
            if not isinstance(docid2, (tuple,list)):
                # docid2 is a single doc
                docid2=[docid2,]

            m=len(docid2)

            for ii in range(m):
                docii=docid2[ii]
                for jj in range(n):
                    docjj=docids1[jj]
                    if docii!=docjj:
                        job_list.append((ii, docii, docjj, self.meta_dict[docii],
                            self.meta_dict[docjj]))

        return 0,jobid,job_list


    @pyqtSlot()
    def jobListReady(self):
        '''After getting all jobs, launch threads and process jobs
        '''

        rec,jobid,job_list=self.master1.results[0]
        LOGGER.debug('rec from job list prepare = %s' %rec)

        if rec==0 and len(job_list)>0:
            self.parent.progressbar.setMaximum(0)
            self.parent.progressbar.setVisible(True)
            # make a separate master for fuzzy matching. This separation of
            # joblist-preparing and fuzzy matching is for easy aborting.
            # For large data size, it might take a few seconds to get the job
            # list prepared, therefore 2 threaded calls.
            self.master2=Master(fuzzyMatch,job_list,1,self.parent.progressbar,
                    'busy',self.parent.status_bar,'Computing Fuzzy Matching...')
            self.master2.all_done_signal.connect(self.collectResults)
            self.clear_duplicate_button.clicked.connect(self.master2.abortJobs)
            self.master2.run()

        return


    def collectResults(self):
        '''Collect matching results'''

        new=self.master2.results
        for recii,jobidii,(kii,vii) in new:
            self.scores_dict[kii]=vii

        LOGGER.info('Duplicate search results collected.')
        self.addResultToTree()

        return


    def addResultToTree(self):
        '''Display results'''

        hi_color=self.settings.value('display/folder/highlight_color_br',
                QBrush)

        def createEntry(docid,gid,score):

            meta=self.meta_dict[docid]
            item=QtWidgets.QTreeWidgetItem([
                gid,
                ', '.join(meta['authors_l']),
                meta['title'],
                meta['publication'],
                str(meta['year']),
                score,
                str(docid)
                ])

            return item

        edges=[kk for kk,vv in self.scores_dict.items() if vv>=self.min_score]
        # if no duplicates, return
        if len(edges)==0:
            self.noDupLabel.setVisible(True)
            LOGGER.info('No duplicate found.')
            return

        self.noDupLabel.setVisible(False)

        #--------------------Get groups--------------------
        groups={}

        if self.docid2 is None:

            # get connected components
            comps=dfsCC(edges)
            LOGGER.debug('connected components = %s' %comps)

            for ii,cii in enumerate(comps):
                cii.sort()
                groups[cii[0]]=cii[1:]
        else:
            if not isinstance(self.docid2, (tuple,list)):
                # docid2 is a single doc
                docid2=[self.docid2,]
            else:
                docid2=self.docid2

            for ii,docii in enumerate(docid2):
                docs=[kk[1] for kk in edges if kk[0]==docii]
                if len(docs)>0:
                    groups[docii]=docs

        #--------------------Add items--------------------
        # sort by group length
        headers=sorted(groups.keys(), key=lambda x: len(groups[x]), reverse=True)

        for ii, docii in enumerate(headers):

            itemii=createEntry(docii, str(ii+1), '')

            # highlight group header
            for jj in range(self.tree.columnCount()):
                itemii.setBackground(jj, hi_color)

            self.tree.addTopLevelItem(itemii)

            # sort group members by scores
            members=groups[docii]
            scores=[]
            for djj in members:
                if (docii, djj) in edges:
                    sii=self.scores_dict[(docii, djj)]
                else:
                    sii=self.scores_dict[(djj, docii)]
                scores.append(sii)
            members=[x for _,x in sorted(zip(scores,members), reverse=True)]
            scores=sorted(scores,reverse=True)

            # add group members
            for jj,docjj in enumerate(members):
                itemjj=createEntry(docjj, '', str(scores[jj]))
                itemii.addChild(itemjj)

            self.tree.expandAll()

        color=hi_color.color().getRgb()
        color_str='rgb(%s)' %','.join(map(str,color))

        self.tree.setStyleSheet('''
        QTreeWidget::item:has-children { border-left: 1px solid black;
        background-color: %s;}
        ''' %color_str)

        LOGGER.info('Duplicate search results added.')

        return


    def docTreeMenu(self,pos):
        '''Right click menu in the document tree'''

        menu=QtWidgets.QMenu()

        foldername,folderid=self.current_folder
        if folderid=='-1':
            menu.addAction('Delete From Library')
        else:
            menu.addAction('Delete From Current Folder')

        action=menu.exec_(QCursor.pos())

        if action:
            self.delDocs()

        return


    @pyqtSlot()
    def delDocs(self):

        LOGGER.debug('current_folder = %s' %str(self.current_folder))
        foldername,folderid=self.current_folder
        sel_rows=self.tree.selectedItems()

        if len(sel_rows)>0:

            docids=[int(ii.data(6,0)) for ii in sel_rows]
            LOGGER.debug('Selected docids = %s.' %docids)

            if folderid=='-1':
                LOGGER.info('Emit signal for doc deletion in All')
                self.del_doc_from_lib_signal.emit(docids,False)
            else:
                LOGGER.info('Emit signal for doc deletion in folder %s' %foldername)
                self.del_doc_from_folder_signal.emit(docids, foldername,
                        folderid, False)

            for itemii in sel_rows:
                self.tree.invisibleRootItem().removeChild(itemii)

        return



'''
Handles data updating, including updates of the in-memory data dictionary from
editings in  the meta data tabs and from DOI querying; adding docs to folders,
and saving the in-memory data to sqlite database.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from datetime import datetime
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QBrush
from PyQt5 import QtWidgets
from lib import sqlitedb
from lib import widgets



class MainFrameDataSlots:


    def threadedFuncCall2(self,func,joblist,show_message='',max_threads=4,
            get_results=False,close_on_finish=True,progressbar_style='classic'):
        '''Call function in another thread.
        See threadrun_dialog.py for more details
        NOT in use any more.
        '''

        thread_run_dialog=widgets.ThreadRunDialog(func,joblist,
                show_message,max_threads,get_results,close_on_finish,
                progressbar_style,parent=self)
        thread_run_dialog.exec_()
        if get_results:
            return thread_run_dialog.results
        else:
            return



    #######################################################################
    #                      Meta data update functions                      #
    #######################################################################

    def updateTableData(self,docid,meta_dict,field_list=None):
        """Update the in-memory dictionary self.meta_dict

        Args:
            docid (int): id of doc to update.
            meta_dict (DocMeta dict): dict containing the updated meta data for
                                      doc with id = <docid>.
        Kwargs:
            field_list (None or list): if list of str, giving the keys in
                                       <meta_dict> that needs update.
                                       if None, adding a new doc (<docid> not in
                                       self.meta_dict).

        Returns:
            docid (int): id assigned to added doc. If <docid> was None, this
                         gives the id assigned to it. Otherwise it return
                         <docid>.
        """

        if docid is None:

            if len(self.meta_dict)==0:
                docid=1
            else:
                docid=max(self.meta_dict.keys())+1

            self.logger.info('Add new doc. Given id=%s' %docid)

            # update folder_data
            foldername,folderid=self._current_folder
            if folderid not in ['-1', '-2', '-3']:
                self.folder_data[folderid].append(docid)

                if (folderid, foldername) not in meta_dict['folders_l']:
                    meta_dict['folders_l'].append((folderid,foldername))
                    self.logger.debug('Add new doc to folder %s. meta_dict["folders_l"]=%s'\
                            %(foldername, meta_dict['folders_l']))

            # update meta_dict
            self.meta_dict[docid]=meta_dict
            # add to needs review folder
            self.folder_data['-2'].append(docid)
            # scroll to and select row in doc table
            self.doc_table.scrollToBottom()
            self.loadDocTable(docids=self._current_docids+[docid,],sel_row=None)

            #xx=self.doc_table.model().rowCount(None)
            # NOTE that the below method may not work when table was empty before
            # adding, as I connected doc_table.currentChanged to selDoc.
            # When table was empty, the index for previous current isnt defined
            # UPDATE: never mind the above.
            self.doc_table.selectRow(self.doc_table.model().rowCount(None)-1)

        else:
            if docid in self.meta_dict:
                self.logger.info('Updating existing doc. docid=%s' %docid)

                for kk in field_list:
                    if kk=='authors_l':
                        self.meta_dict[docid]['firstNames_l']=meta_dict['firstNames_l']
                        self.meta_dict[docid]['lastName_l']=meta_dict['lastName_l']
                    else:
                        self.meta_dict[docid][kk]=meta_dict[kk]
            else:
                self.logger.debug('wtf?')

                self.meta_dict[docid]=meta_dict

            # reload doc table
            self.loadDocTable(docids=self._current_docids,
                    sel_row=self.doc_table.currentIndex().row())

        self.changed_doc_ids.append(docid)

        return docid


    @pyqtSlot(sqlitedb.DocMeta)
    def updateByDOI(self, meta_dict):
        """update in-memory dictionary self.meta_dict via doi query

        args:
            meta_dict (docmeta dict): new dict containing meta data from doi
                                      query.
        """

        docid=self._current_doc
        self.logger.info('Update doc %s by doi' %docid)

        if docid:
            self.meta_dict[docid]=meta_dict
            self.changed_doc_ids.append(docid)
            self.loadDocTable(docids=self._current_docids,
                    sel_row=self.doc_table.currentIndex().row())

        return


    @pyqtSlot()
    def updateNotes(self, docid, note_text):
        """update notes in the in-memory dictionary

        args:
            docid (int): id of doc to update.
            note_text (str): new note texts.
        """

        if docid is None:
            return

        self.meta_dict[docid]['notes']=note_text
        self.changed_doc_ids.append(docid)
        self.logger.info('New notes for docid=%s: %s' %(docid,note_text))

        return


    @pyqtSlot(int,str)
    def addDocToFolder(self, docid, folderid):
        """Add a doc to a folder

        Args:
            docid (int): id of doc.
            folderid (str): id of folder to accept doc.
        """

        self.logger.info('docid=%s, folderid=%s' %(docid,folderid))

        #----------Add folder to doc's folders_l----------
        docfolders=self.meta_dict[docid]['folders_l']
        # note folderid here is an int
        newfolder=(int(folderid), self.folder_dict[folderid][0])
        if newfolder not in docfolders:
            docfolders.append(newfolder)
            self.meta_dict[docid]['folders_l']=docfolders

        #-------------Add docid to folder_data-------------
        if docid not in self.folder_data[folderid]:
            self.folder_data[folderid].append(docid)

        self.logger.debug('Updated meta_dict["folders_l"] = %s'\
                %self.meta_dict[docid]['folders_l'])
        self.logger.debug('Updated folder_data = %s' %self.folder_data[folderid])

        #-------------Restoring a trashed doc-------------
        current_folderid=self._current_folder[1]
        trashed_folders=self._trashed_folder_ids+['-3']
        if current_folderid in trashed_folders and folderid not in trashed_folders:
            self.logger.info('Restoring a trashed doc.')
            self.logger.debug('Updated deletionPending = %s'\
                    %self.meta_dict[docid]['deletionPending'])
            self.meta_dict[docid]['deletionPending']=='false'

            # remove doc from current folder when restoring
            if docid in self.folder_data[current_folderid]:
                self.folder_data[current_folderid].remove(docid)
                self.loadDocTable(folder=self._current_folder,sel_row=None,sortidx=4)

        # add highlight to folder
        hi_color=self.settings.value('display/folder/highlight_color_br',
                QBrush)

        mii=self.libtree.findItems(folderid, Qt.MatchExactly | Qt.MatchRecursive,
                column=1)
        if len(mii)>0:
            for mjj in mii:
                mjj.setBackground(0, hi_color)

        self.changed_doc_ids.append(docid)


        return


    @pyqtSlot()
    def saveToDatabase(self):
        """Save in-memory data to sqlite file

        self.changed_folder_ids contains ids of folder to update.
        self.changed_doc_ids contains ids of docs to update.
        Clear these two after saving.
        """

        mtime=datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        self.logger.info('Save called. %s' %mtime)

        if len(self.changed_folder_ids)==0 and len(self.changed_doc_ids)==0:
            return

        # remove duplicates
        self.changed_folder_ids=list(set(self.changed_folder_ids))
        self.changed_doc_ids=list(set(self.changed_doc_ids))

        self.logger.debug('Folders to save: %s' %self.changed_folder_ids)
        self.logger.debug('Docs to save: %s' %self.changed_doc_ids)

        self.status_bar.setVisible(True)
        self.status_bar.showMessage('Saving. Please standby ...')
        # progressbar doesn't work in the main thread?
        self.progressbar.setVisible(True)
        self.progressbar.setMaximum(0)
        QtWidgets.QApplication.processEvents()

        #----------------Save folders first----------------
        if len(self.changed_folder_ids)>0:

            sqlitedb.saveFoldersToDatabase(self.db,
                    self.changed_folder_ids, self.folder_dict,
                    self.folder_data,
                    self.settings.value('saving/current_lib_folder'))

            self.changed_folder_ids=[]
            self.logger.info('Folder changes saved to database.')

        #--------------------Save docs--------------------
        for docid in self.changed_doc_ids:
            self.logger.info('Saving doc %s' %docid)
            sqlitedb.metaDictToDatabase(self.db,docid,
                    self.meta_dict.get(docid),
                    self.settings.value('saving/current_lib_folder'),
                    self.settings.value('saving/rename_files'))

        self.changed_doc_ids=[]
        self.settings.sync()
        self.status_bar.clearMessage()
        self.progressbar.setVisible(False)

        self.logger.info('Saving completed.')

        return


    @pyqtSlot()
    def createFailFolder(self, show_text, docids):
        """Show in the doc table docs in a failed task

        Args:
            show_text (str): info texts to show in the label describing the task.
            docids (list): list of doc ids (int) to load in the doc table.

        NOTE: this is a misnomer, no folder is created. rename later
        """

        self.clear_filter_label.setText('Failed tasks in %s' %show_text)
        self.clear_filter_frame.setVisible(True)
        self.loadDocTable(docids=docids)

        return








'''
Perform searches in sqlite database.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import sqlite3
import logging

LOGGER=logging.getLogger(__name__)



#######################################################################
#                             fts5 search NOT IN USE                  #
#######################################################################


def searchTitle(db, text):

    cin=db.cursor()

    query='''SELECT rowid,
    snippet(Documents, 14, '<b>', '</b>', '<b>...</b>', 5) title
    FROM Documents
    WHERE (Documents.title MATCH ?)
    ORDER BY rank
    '''
    print(query)

    ret=cin.execute(query,('%s*' %text,))
    ret=ret.fetchall()
    for ii in ret:
        print(ii)

    return ret

def searchAuthors(db, text):
    cin=db.cursor()

    query='''SELECT did, ("firstNames" || " " || "lastName") as name
    FROM DocumentContributors
    WHERE (DocumentContributors MATCH ?)
    ORDER By rank
    '''

    print(query)
    ret=cin.execute(query, ('%s*' %text,))
    ret=ret.fetchall()

    for ii in ret:
        print(ii)

    return ret

def searchKeywords(db, text):
    cin=db.cursor()

    query='''SELECT did, text
    FROM DocumentKeywords
    WHERE (DocumentKeywords.text MATCH ?)
    ORDER By rank
    '''

    print(query)
    ret=cin.execute(query, ('%s*' %text,))
    ret=ret.fetchall()

    for ii in ret:
        print(ii)

    return ret

def searchTags(db, text):
    cin=db.cursor()

    query='''SELECT did, tag
    FROM DocumentTags
    WHERE (DocumentTags.tag MATCH ?)
    ORDER By rank
    '''

    print(query)
    ret=cin.execute(query, ('%s*' %text,))
    ret=ret.fetchall()

    for ii in ret:
        print(ii)

    return ret

def searchNotes(db, text):
    cin=db.cursor()

    query='''SELECT did,
    snippet(DocumentNotes, 1, '<b>', '</b>', '<b>...</b>', 6) note
    FROM DocumentNotes
    WHERE (DocumentNotes.note MATCH ?)
    ORDER By rank
    '''

    print(query)
    ret=cin.execute(query, ('%s*' %text,))
    ret=ret.fetchall()

    for ii in ret:
        print(ii)

    return ret

def searchMultiple(db, text, field_list, folderid):

    cin=db.cursor()

    #-------------Append * to search term-------------
    text='%s*' %text

    #----------------Create temp table----------------
    query='''
    DROP TABLE IF EXISTS temp.search_res'''
    cin.execute(query)

    query='''
    CREATE VIRTUAL TABLE temp.search_res USING fts5(
    did, text, _rank)
    '''
    cin.execute(query)

    #---------------Compose search query---------------
    queries=[]
    for kk in field_list:
        if kk=='Authors':
            qkk='''SELECT did, ("firstNames" || " " || "lastName") as name, rank
            FROM DocumentContributors
            WHERE DocumentContributors MATCH ?
            '''
        elif kk=='Title':
            qkk='''SELECT rowid, Documents.title, rank
            FROM Documents
            WHERE Documents.title MATCH ?'''
        elif kk=='Keywords':
            qkk='''SELECT did, text, rank
            FROM DocumentKeywords
            WHERE DocumentKeywords.text MATCH ?'''
        elif kk=='Tags':
            qkk='''SELECT did, tag, rank
            FROM DocumentTags
            WHERE DocumentTags.tag MATCH ?'''
        elif kk=='Notes':
            qkk='''SELECT did,
            snippet(DocumentNotes, 1, '<b>', '</b>', '<b>...</b>', 6) note,
            rank
            FROM DocumentNotes
            WHERE DocumentNotes.note MATCH ?'''

        queries.append(qkk)

    query='''INSERT INTO temp.search_res (did, text, _rank)
    %s
    ORDER BY rank
    ''' %('UNION ALL\n'.join(queries))
    print(query)

    ret=cin.execute(query, (text,)*len(queries))

    #---------Get results and filter by folder---------
    if folderid=='-1':
        query='''SELECT search_res.did, search_res.text, search_res._rank
        FROM search_res'''
        ret=cin.execute(query)

    elif folderid=='-2':
        query='''SELECT search_res.did, search_res.text, search_res._rank
        FROM search_res
        LEFT JOIN Documents ON Documents.rowid=search_res.did
        WHERE Documents.confirmed='false'
        '''
        ret=cin.execute(query)

    elif folderid=='-3':
        query='''SELECT search_res.did, search_res.text, search_res._rank
        FROM search_res
        LEFT JOIN Documents ON Documents.rowid=search_res.did
        WHERE Documents.deletionPending='true'
        '''
        ret=cin.execute(query)

    else:
        query='''SELECT search_res.did, search_res.text, search_res._rank
        FROM search_res
        LEFT JOIN DocumentFolders ON DocumentFolders.did=search_res.did
        WHERE DocumentFolders.folderid=?'''
        ret=cin.execute(query, (folderid,))

    print(query)
    ret=ret.fetchall()
    for ii in ret:
        print(ii)

    return ret

def getFolders(db):

    #-----------------Get all folders-----------------
    query='''SELECT id, name, parentId
    FROM Folders
    '''
    ret=db.execute(query)
    data=ret.fetchall()

    # dict, key: folderid, value: (folder_name, parent_id)
    # note: convert id to str
    df=dict([(str(ii[0]), (ii[1], str(ii[2]))) for ii in data])

    # add Default
    #df['0']=('Default', '-1')

    return df


def getSubFolders(folder_dict, folderid):

    results=[kk for kk,vv in folder_dict.items() if vv[1]==folderid]

    subs=[]
    for fii in results:
        subids=getSubFolders(folder_dict,fii)
        subs.extend(subids)

    results.extend(subs)

    return results


#######################################################################
#             LIKE search ONLY searchMultipleLike2 IN USE             #
#######################################################################

def searchTitleLike(db, text, folderid, desend=False):

    cin=db.cursor()

    query='''SELECT Documents.id, Documents.title
    FROM Documents
    LEFT JOIN DocumentFolders
    ON DocumentFolders.did=Documents.id
    WHERE (Documents.title LIKE ?) AND
    %s
    ORDER BY Documents.title
    '''

    if desend:
        folder_dict=getFolders(db)
        subfolderids=getSubFolders(folder_dict, folderid)+[folderid,]
        print('# <searchTitleLike>: subfolderids=',subfolderids)
        folder_str='(DocumentFolders.folderid IN (%s))' %','.join(['?']*len(subfolderids))
        folder_value=tuple(subfolderids)
    else:
        folder_str='(DocumentFolders.folderid = ?)'
        folder_value=(folderid,)

    ret=cin.execute(query %folder_str,('%%%s%%' %text,) + folder_value)
    ret=ret.fetchall()

    return ret

def searchAuthorsLike(db, text):

    cin=db.cursor()
    query='''SELECT did, ("firstNames" || " " || "lastName") as name
    FROM DocumentContributors
    WHERE (name LIKE ?)
    ORDER By name
    '''
    ret=cin.execute(query, ('%%%s%%' %text,))
    ret=ret.fetchall()

    return ret

def searchKeywordsLike(db, text):

    cin=db.cursor()
    query='''SELECT did, text
    FROM DocumentKeywords
    WHERE (DocumentKeywords.text LIKE ?)
    ORDER By text
    '''
    ret=cin.execute(query, ('%%%s%%' %text,))
    ret=ret.fetchall()

    return ret



def searchTagsLike(db, text):

    cin=db.cursor()
    query='''SELECT did, tag
    FROM DocumentTags
    WHERE (DocumentTags.tag LIKE ?)
    ORDER By tag
    '''
    ret=cin.execute(query, ('%%%s%%' %text,))
    ret=ret.fetchall()

    return ret


def searchNotesLike(db, text):

    cin=db.cursor()
    query='''SELECT did, note
    FROM DocumentNotes
    WHERE (DocumentNotes.note LIKE ?)
    ORDER By note
    '''

    ret=cin.execute(query, ('%%%s%%' %text,))
    ret=ret.fetchall()

    return ret


def searchMultipleLike(db, text, field_list, folderid, desend=False):

    cin=db.cursor()

    #-------------Append % to search term-------------
    text='%%%s%%' %text

    #----------------Create temp table----------------
    query='''
    DROP TABLE IF EXISTS temp.search_res'''
    cin.execute(query)

    query='''
    CREATE TABLE temp.search_res (
    did, text, field)
    '''
    cin.execute(query)

    #---------------Compose search query---------------
    queries=[]
    for kk in field_list:
        if kk=='Authors':
            qkk='''SELECT did, ("firstNames" || " " || "lastName") as name, 'authors'
            FROM DocumentContributors
            WHERE name LIKE ?'''
        elif kk=='Title':
            qkk='''SELECT id, Documents.title, 'title'
            FROM Documents
            WHERE Documents.title LIKE ?'''
        elif kk=='Keywords':
            qkk='''SELECT did, text, 'keywords'
            FROM DocumentKeywords
            WHERE DocumentKeywords.text LIKE ?'''
        elif kk=='Tags':
            qkk='''SELECT did, tag, 'tag'
            FROM DocumentTags
            WHERE DocumentTags.tag LIKE ?'''
        elif kk=='Notes':
            qkk='''SELECT did, note, 'note'
            FROM DocumentNotes
            WHERE DocumentNotes.note LIKE ?'''
        elif kk=='Publication':
            qkk='''SELECT id, Documents.publication, 'publication'
            FROM Documents
            WHERE Documents.publication LIKE ?'''
        elif kk=='Abstract':
            qkk='''SELECT id, Documents.abstract, 'abstract'
            FROM Documents
            WHERE Documents.abstract LIKE ?'''
        else:
            continue

        queries.append(qkk)

    query='''INSERT INTO temp.search_res (did, text, field)
    %s
    ''' %('UNION ALL\n'.join(queries))

    ret=cin.execute(query, (text,)*len(queries))

    #---------Get results and filter by folder---------
    query='''SELECT search_res.did,
    group_concat(search_res.field, ',')
    FROM search_res
    %s
    GROUP BY search_res.did
    '''
    #query='''SELECT search_res.did,
    #search_res.text,
    #search_res.field
    #FROM search_res
    #%s
    #ORDER BY search_res.did
    #'''

    if folderid=='-1':
        query=query %''
        ret=cin.execute(query)

    elif folderid=='-2':
        query=query %'''
        LEFT JOIN Documents ON Documents.id=search_res.did
        WHERE Documents.confirmed='false'
        '''
        ret=cin.execute(query)

    elif folderid=='-3':
        query=query %'''
        LEFT JOIN Documents ON Documents.id=search_res.did
        WHERE Documents.deletionPending='true'
        '''
        ret=cin.execute(query)

    else:
        query=query %'''
        LEFT JOIN DocumentFolders ON DocumentFolders.did=search_res.did
        WHERE DocumentFolders.folderid=?
        '''
        ret=cin.execute(query, (folderid,))

    #print(query)
    ret=ret.fetchall()

    #for ii in ret:
        #print(ii)

    return ret

def searchMultipleLike2(db, text, field_list, folderid, desend=False):
    """Combined sqlite searches in multiple tables

    Args:
        db (sqlite connection): sqlite connection.
        text (str): search text.
        field_list (list): list of fields to search, including 'Authors',
                          'Title', 'Keywords', 'Tags', 'Notes', 'Publication',
                          'Abstract'.
        folderid (str): id of folder, search in done within docs in this folder.

    Kwargs:
        desend (bool): whether to include subfolders (by walking done folder
                       tree) of given folder.

    Returns: rec (list): list of doc ids matching search, together with the
        field names where the match is found. E.g.

        [(1, 'authors,title'),
         (10, 'title,keywords'),
         (214, 'abstract,tag'),
         ...
        ]
    """

    cin=db.cursor()

    #-------------Append % to search term-------------
    text='%%%s%%' %text

    #----------------Create temp table----------------
    query='''
    DROP TABLE IF EXISTS temp.search_res'''
    cin.execute(query)

    query='''
    CREATE TABLE temp.search_res (
    did, text, field)
    '''
    cin.execute(query)

    #---------------Compose search query---------------
    queries=[]
    for kk in field_list:
        if kk=='Authors':
            qkk='''SELECT did, ("firstNames" || " " || "lastName") as name, 'authors'
            FROM DocumentContributors
            WHERE name LIKE ?'''
        elif kk=='Title':
            qkk='''SELECT id, Documents.title, 'title'
            FROM Documents
            WHERE Documents.title LIKE ?'''
        elif kk=='Keywords':
            qkk='''SELECT did, text, 'keywords'
            FROM DocumentKeywords
            WHERE DocumentKeywords.text LIKE ?'''
        elif kk=='Tags':
            qkk='''SELECT did, tag, 'tag'
            FROM DocumentTags
            WHERE DocumentTags.tag LIKE ?'''
        elif kk=='Notes':
            qkk='''SELECT did, note, 'note'
            FROM DocumentNotes
            WHERE DocumentNotes.note LIKE ?'''
        elif kk=='Publication':
            qkk='''SELECT id, Documents.publication, 'publication'
            FROM Documents
            WHERE Documents.publication LIKE ?'''
        elif kk=='Abstract':
            qkk='''SELECT id, Documents.abstract, 'abstract'
            FROM Documents
            WHERE Documents.abstract LIKE ?'''
        else:
            continue

        queries.append(qkk)

    query='''INSERT INTO temp.search_res (did, text, field)
    %s
    ''' %('UNION ALL\n'.join(queries))

    ret=cin.execute(query, (text,)*len(queries))

    #---------Get results and filter by folder---------
    query='''SELECT search_res.did,
    group_concat(search_res.field, ',')
    FROM search_res
    %s
    GROUP BY search_res.did
    '''

    #---------Compose folder filtering string---------
    if folderid=='-1':
        query=query %''
        ret=cin.execute(query)
    elif folderid=='-2':
        query=query %'''
        LEFT JOIN Documents ON Documents.id=search_res.did
        WHERE Documents.confirmed='false'
        '''
        ret=cin.execute(query)
    else:
        if desend:
            folder_dict=getFolders(db)
            subfolderids=getSubFolders(folder_dict, folderid)+[folderid,]
        else:
            subfolderids=[folderid,]

        LOGGER.debug('subfolder ids = %s' %subfolderids)

        query=query %'''
        LEFT JOIN DocumentFolders ON DocumentFolders.did=search_res.did
        WHERE (DocumentFolders.folderid IN (%s))
        ''' %','.join(['?']*len(subfolderids))
        ret=cin.execute(query, subfolderids)

    ret=ret.fetchall()

    return ret





if __name__=='__main__':

    dbfile='../nonfts3.sqlite'

    db = sqlite3.connect(dbfile)
    #cin=db.cursor()

    #aa=searchTitle(db, 'effect')
    #aa=searchAuthors(db, 'Che')
    #aa=searchKeywords(db, 'ENSO')
    #aa=searchTags(db, 'ENSO')
    #aa=searchNotes(db, 'the')
    #aa=searchMultiple(db, 'ENSO')

    field_list=['Authors', 'Title', 'Keywords', 'Tag', \
            'Abstract', 'Publication', 'Notes']
    #aa=searchMultipleLike2(db, 'ENSO', field_list, '16')
    aa=searchMultipleLike2(db, 'ENSO', ['Title',' Keywords', 'Abstract'], '16', True)
    bb=searchTitleLike(db, 'ENSO', '16', True)
    for ii in bb:
        print(ii)

    #folder_dict=getFolders(db)
    #ff=getSubFolders(folder_dict,'16')




'''
Import dialog.

MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import shutil
import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QSize
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QDialogButtonBox, QStyle
from ..tools import getHLine
from .threadrun_dialog import ThreadRunDialog
#from import_mendeley import importMendeley
from .. import import_mendeley

LOGGER=logging.getLogger(__name__)


class ResultDialog(QtWidgets.QDialog):

    def __init__(self,main_text='',info_text='',detailed_text='',parent=None):
        super(self.__class__,self).__init__(parent=parent)

        self.main_text=main_text
        self.info_text=info_text
        self.detailed_text=detailed_text

        self.setWindowTitle('Import Result')
        self.resize(450,250)
        self.grid=QtWidgets.QGridLayout(self)

        icon=self.style().standardIcon(QStyle.SP_MessageBoxInformation)
        icon_label=QtWidgets.QLabel()
        icon_label.setPixmap(icon.pixmap(QSize(32,32)))
        self.grid.addWidget(icon_label,0,0)

        self.main_text_label=QtWidgets.QLabel(main_text)
        label_font=QFont('Serif',12,QFont.Bold)
        self.main_text_label.setFont(label_font)
        self.grid.addWidget(self.main_text_label,0,1)

        self.info_text_label=QtWidgets.QLabel(info_text)
        self.info_text_label.setTextFormat(Qt.RichText)
        self.info_text_label.setWordWrap(True)
        self.grid.addWidget(self.info_text_label,1,1,1,2)

        self.open_lib_checkbox=QtWidgets.QCheckBox('Open Imported Library?')
        self.open_lib_checkbox.setChecked(True)
        self.grid.addWidget(self.open_lib_checkbox,2,0)

        self.buttons=QDialogButtonBox(self)

        self.ok_button=QtWidgets.QPushButton('Ok')
        self.ok_button.setDefault(True)
        self.ok_button.setAutoDefault(True)

        self.detail_button=QtWidgets.QPushButton('Show Details')
        self.detail_button.setDefault(False)
        self.detail_button.setAutoDefault(False)
        self.detail_button.clicked.connect(self.detailButtonClicked)

        #self.create_folder_button=QtWidgets.QPushButton('Show Failed Docs')
        #self.create_folder_button.setDefault(False)
        #self.create_folder_button.setAutoDefault(False)
        #self.create_folder_button.clicked.connect(self.createFailList)

        self.buttons.addButton(self.ok_button,QDialogButtonBox.AcceptRole)
        self.buttons.addButton(self.detail_button,QDialogButtonBox.ActionRole)
        #self.buttons.addButton(self.create_folder_button,
                #QDialogButtonBox.ActionRole)

        self.grid.addWidget(self.buttons,3,2)

        self.text_edit=QtWidgets.QTextEdit(self)
        self.text_edit.setText(detailed_text)
        self.text_edit.setVisible(False)

        self.grid.addWidget(self.text_edit,4,0,1,3)

        self.buttons.accepted.connect(self.accept)

        #self.exec_()

    def setText(self,text):
        self.main_text=text
        self.main_text_label.setText(text)
        return

    def setInformativeText(self,text):
        self.info_text=text
        self.info_text_label.setText(text)
        return

    def setDetailedText(self,text):
        self.detailed_text=text
        self.text_edit.setText(text)
        return


    def detailButtonClicked(self):
        if self.detailed_text=='':
            return

        if self.text_edit.isVisible():
            self.detail_button.setText('Show Details')
            self.text_edit.setVisible(False)
        else:
            self.detail_button.setText('Hide Details')
            self.text_edit.setVisible(True)

        return

    #def createFailList(self):
        #self.create_fail_summary.emit()
        #return

    def accept(self):
        if self.open_lib_checkbox.isChecked():
            print('# <accept>: open lib')
            #self.open_lib_signal.emit()
            super(self.__class__, self).accept()
        else:
            super(self.__class__, self).reject()



class ImportDialog(QtWidgets.QDialog):

    open_lib_signal=pyqtSignal(str)  # sqlite file name

    def __init__(self,settings,parent):
        '''
        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py
        '''

        super(ImportDialog,self).__init__(parent=parent)

        self.settings=settings
        self.parent=parent

        self.label_color='color: rgb(0,0,140); background-color: rgb(235,235,240)'
        self.title_label_font=QFont('Serif',12,QFont.Bold)
        self.sub_title_label_font=QFont('Serif',10,QFont.Bold)

        self.resize(900,600)
        self.setWindowTitle('Bulk Import')
        self.setWindowModality(Qt.ApplicationModal)

        v_layout=QtWidgets.QVBoxLayout()
        h_layout=QtWidgets.QHBoxLayout()
        #h_layout.setContentsMargins(10,40,10,20)
        self.setLayout(v_layout)

        title_label=QtWidgets.QLabel('    Choose Import Type')
        title_label.setFont(QFont('Serif',12,QFont.Bold))
        v_layout.addWidget(title_label)

        v_layout.addLayout(h_layout)

        self.cate_list=QtWidgets.QListWidget(self)
        #self.list.setSizePolicy(getXMinYExpandSizePolicy())
        self.cate_list.setMaximumWidth(200)
        h_layout.addWidget(self.cate_list)

        #self.cate_list.setStyleSheet('''
            #QListWidget::item { border: 0px solid rgb(235,235,240);
            #font: 14px;
            #background-color: rgb(205,205,245);
            #color: rgb(100,10,13) };
            #background-color: rgb(230,234,235);
            #''')

        self.cate_list.addItems(['Import From Mendeley', 'Import From Zotero',
            'Import From EndNote'])

        self.content_vlayout=QtWidgets.QVBoxLayout()
        h_layout.addLayout(self.content_vlayout)

        self.buttons=QDialogButtonBox(QDialogButtonBox.Close,
            Qt.Horizontal, self)
        self.import_button=self.buttons.addButton('Import',
                QDialogButtonBox.ApplyRole)

        self.import_button.clicked.connect(self.doImport)
        self.buttons.rejected.connect(self.reject)

        self.content_vlayout.addWidget(self.buttons)

        self.cate_list.currentItemChanged.connect(self.cateSelected)
        self.cate_list.setCurrentRow(0)



    @pyqtSlot(QtWidgets.QListWidgetItem)
    def cateSelected(self,item):
        '''Load widgets for a selected category

        Args:
            item (QListWidgetItem): selected category item.
        '''

        item_text=item.text()
        LOGGER.debug('item.text() = %s' %item_text)

        if self.content_vlayout.count()>1:
            self.content_vlayout.removeWidget(self.content_frame)

        if item_text=='Import From Mendeley':
            self.content_frame=self.loadImportMendeley()
        elif item_text=='Import From Zotero':
            self.content_frame=self.loadImportZotero()
        elif item_text=='Import From EndNote':
            self.content_frame=self.loadImportEndNote()

        self.content_vlayout.insertWidget(0,self.content_frame)

        return


    def createFrame(self, title):
        '''Create a template frame for a category page

        Args:
            title (str): title of the category

        Returns:
            scroll (QScrollArea): a scroll area.
            va (QVBoxLayout): the vertical box layout used in scroll.
        '''

        frame=QtWidgets.QWidget(self)
        scroll=QtWidgets.QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setWidget(frame)
        va=QtWidgets.QVBoxLayout()
        frame.setLayout(va)
        va.setSpacing(int(va.spacing()*2))

        label=QtWidgets.QLabel(title)
        label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)
        va.addWidget(label)
        #va.addWidget(getHLine(self))

        return scroll, va


    def loadImportMendeley(self):
        '''Load Mendeley import category page'''

        scroll,va=self.createFrame('Import From Mendeley')
        self.current_task='mendeley'

        label=QtWidgets.QLabel('(Notice: Only Mendeley version < 1.19 can be imported. Later version of Mendeley encrypts the database file.)')
        label.setStyleSheet('Font: bold')
        label.setWordWrap(True)
        va.addWidget(label)

        #---------------------Lib name---------------------
        '''
        label=QtWidgets.QLabel('Name your library')
        va.addWidget(label)

        self.lib_name_le=QtWidgets.QLineEdit()
        regex=QRegExp("[a-z-A-Z_\d]+")
        validator = QRegExpValidator(regex)
        self.lib_name_le.setValidator(validator)

        ha=QtWidgets.QHBoxLayout()
        ha.addWidget(self.lib_name_le)

        label=QtWidgets.QLabel('(Only alphanumeric characters and "-", "_" are allowed)')
        ha.addWidget(label)
        va.addLayout(ha)
        '''
        #-----------------New sqlite file-----------------
        label=QtWidgets.QLabel('Name your library')
        va.addWidget(label)

        self.lib_name_le=QtWidgets.QLineEdit()
        button=QtWidgets.QPushButton(self)
        button.setText('Open')
        button.clicked.connect(lambda: self.outFileChooseButtonClicked(
            self.lib_name_le))
        ha=QtWidgets.QHBoxLayout()
        ha.addWidget(self.lib_name_le)
        ha.addWidget(button)
        va.addLayout(ha)

        va.addWidget(getHLine())

        #-----------------Sqlite file sel-----------------
        label=QtWidgets.QLabel('''Select the sqlite database file <br/>
        Default location: <br/>
        <br/>
        * Linux: ~/.local/share/data/Mendeley Ltd./Mendeley Desktop/<your_email@www.mendeley.com.sqlite. <br/>
        '''
        )
        label.setTextFormat(Qt.RichText)
        va.addWidget(label)

        self.mendeley_file_le=QtWidgets.QLineEdit()
        button=QtWidgets.QPushButton(self)
        button.setText('Open')
        button.clicked.connect(lambda: self.importFileChooseButtonClicked(
            self.mendeley_file_le))

        ha=QtWidgets.QHBoxLayout()
        ha.addWidget(self.mendeley_file_le)
        ha.addWidget(button)

        va.addLayout(ha)

        #----------------------Notice----------------------
        va.addWidget(getHLine())
        va.addStretch()

        label=QtWidgets.QLabel('''Notice: Import will try to export the annotations (highlights, notes) you made in the Mendeley library. <br/>
                "Cananical" documents (those don't belong to any folder) will be put to the "Default" folder.''')
        label.setTextFormat(Qt.RichText)
        va.addWidget(label)


        return scroll


    @pyqtSlot(QtWidgets.QLineEdit)
    def outFileChooseButtonClicked(self, le):
        '''Prompt to select output sqlite file name and set it to lineedit'''

        storage_folder=self.settings.value('saving/storage_folder',str)
        fname = QtWidgets.QFileDialog.getSaveFileName(self,
                'Name your sqlite database file',
                storage_folder,
                "sqlite Files (*.sqlite);; All files (*)")[0]

        if fname:
            # make sure has .sqlite ext
            dirname,filename=os.path.split(fname)
            lib_name,ext=os.path.splitext(filename)
            if ext=='':
                filename='%s.sqlite' %lib_name
                fname=os.path.join(dirname,filename)

            LOGGER.info('Choose file name %s' %fname)
            le.setText(fname)

        return


    @pyqtSlot(QtWidgets.QLineEdit)
    def importFileChooseButtonClicked(self, le):
        '''Prompt to select input sqlite file name and set it to lineedit'''

        fname = QtWidgets.QFileDialog.getOpenFileName(self,
                'Select your Mendeley sqlite database file',
                '',
                "sqlite Files (*.sqlite);; All files (*)")[0]

        if fname:
            LOGGER.info('Choose file name %s' %fname)
            le.setText(fname)

        return


    def loadImportZotero(self):
        '''Load Zotero import category page'''

        scroll,va=self.createFrame('Import From Zotero')
        self.current_task='zotero'

        return scroll


    def loadImportEndNote(self):
        '''Load EndNote import category page'''

        scroll,va=self.createFrame('Import From EndNote')
        self.current_task='endnote'

        return scroll


    def doImport(self):

        LOGGER.info('task = %s' %self.current_task)

        if self.current_task=='mendeley':
            self.doMendeleyImport1()
        elif self.current_task=='zotero':
            pass
        elif self.current_task=='endnote':
            pass

        return


    def doMendeleyImport1(self):
        '''Do Mendeley import, part 1

        This part is responsible for collecting arguments from widgets, and
        feed into importMendeleyPreprocess(). See import_mendeley.py for more
        details.
        '''

        #-----------Get output sqlite file name-----------
        file_out_name=self.lib_name_le.text()
        if file_out_name=='':
            self.popUpGiveName()
            return

        if os.path.exists(file_out_name):
            choice=QtWidgets.QMessageBox.question(self, 'sqlite file already exists',
                    'Overwrite the file %s?' %file_out_name,
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

            if choice==QtWidgets.QMessageBox.Yes:
                os.remove(file_out_name)
            if choice==QtWidgets.QMessageBox.No:
                return

        #------------Get input sqlite file name------------
        file_in_name=self.mendeley_file_le.text()
        if file_in_name=='':
            self.popUpGiveFile()
            return

        if not os.path.exists(file_in_name):
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setWindowTitle('File not found')
            msg.setText("Can't find input file %s" %file_in_name)
            msg.exec_()
            return

        LOGGER.debug('file_in_name = %s' %file_in_name)
        LOGGER.debug('file_out_name = %s' %file_out_name)
        LOGGER.debug('Launching thread...')

        '''
        self.master1=Master(import_mendeley.importMendeleyPreprocess,
                [(0, file_in_name, file_out_name)],
                1, self.parent.main_frame.progressbar,
                'busy', self.parent.main_frame.status_bar,
                'Connecting databases...')

        self.master1.all_done_signal.connect(self.doMendeleyImport2)
        self.master1.run()
        '''
        #------------------Run in thread------------------
        self.thread_run_dialog1=ThreadRunDialog(
                import_mendeley.importMendeleyPreprocess,
                [(0, file_in_name, file_out_name)],
                show_message='Connecting database...',
                max_threads=1,
                get_results=True,
                close_on_finish=True,
                progressbar_style='busy',
                post_process_func=None,
                parent=self)

        self.thread_run_dialog1.master.all_done_signal.connect(
                self.doMendeleyImport2)
        self.thread_run_dialog1.exec_()

        return


    def doMendeleyImport2(self):
        '''Do Mendeley import, part 2

        After connecting to both sqlite databases, this part is responsible for
        copying document data over, and copying attachment files.
        See import_mendeley.importMendeleyCopyData() for more details.
        '''

        #-------------Get results from part 1-------------
        file_out_name=self.lib_name_le.text()
        step1_results=self.thread_run_dialog1.results[0]
        rec, _, dbin, dbout, docids,lib_folder,lib_name=step1_results
        LOGGER.info('return code of importMendeleyPreprocess: %s' %rec)

        if rec==1:
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Error)
            msg.setWindowTitle('Oopsie')
            msg.setText("Failed to process database files.         ")
            msg.exec_()
            LOGGER.warning('Failed to run importMendeleyPreprocess().')

            #----------------Remove sqlite file----------------
            if os.path.exists(file_out_name):
                os.remove(file_out_name)
                LOGGER.info('Remove sqlite database file %s' %file_out_name)
            if os.path.exists(lib_folder):
                shutil.rmtree(lib_folder)
                LOGGER.info('Remove lib folder %s' %lib_folder)

            return

        rename_files=self.settings.value('saving/rename_files', 1)
        LOGGER.debug('rename_files = %s' %rename_files)

        #-----------------Prepare job list-----------------
        self.job_list=[]
        for ii, docii in enumerate(docids):
            self.job_list.append((ii, dbin, dbout, lib_name, lib_folder,
                rename_files, ii, docii))

        # this last job signals a sqlite commit
        self.job_list.append((-1, dbin, dbout, lib_name, lib_folder,
            rename_files, ii, None))

        #------------------Run in thread------------------
        self.thread_run_dialog2=ThreadRunDialog(
                import_mendeley.importMendeleyCopyData,
                self.job_list,
                show_message='Transfering data...',
                max_threads=1,
                get_results=False,
                close_on_finish=False,
                progressbar_style='classic',
                post_process_func=None,
                parent=self)

        self.thread_run_dialog2.master.all_done_signal.connect(
                lambda: self.postImport(file_out_name))
        self.thread_run_dialog2.exec_()

        return


    @pyqtSlot()
    def postImport(self, file_name):
        '''Show feedbacks after Mendeley import

        Args:
            file_name (str): output sqlite file path.
        '''

        #-------------Get results from part 2-------------
        step2_results=self.thread_run_dialog2.master.results
        rec=step2_results[-1][0]  # last step rec
        LOGGER.info('return code of importMendeleyCopyData: %s' %rec)

        #---------Succeeded in committing database---------
        if rec==0:
            fail_list=[]
            pdf_fail_list=[]

            for recii, jobii, fail_fileii in step2_results[:-1]:
                if recii==1:
                    docii=self.job_list[jobii][7]
                    entryii='* docid = %s' %docii
                    LOGGER.warning('Failed to import doc id = %s' %docii)
                    fail_list.append(entryii)
                elif recii==2:
                    entryii='* PDF file(s) = %s' %fail_fileii
                    LOGGER.warning('Failed to export annotated pdf(s) %s'\
                            %fail_fileii)
                    pdf_fail_list.append(entryii)

            #-----------------Show failed jobs-----------------
            if len(fail_list)>0 or len(pdf_fail_list)>0:

                msg=ResultDialog()
                msg.setText('Errors encountered.')

                # only doc transfer failures
                if len(fail_list)>0 and len(pdf_fail_list)==0:
                    msg.setInformativeText('Failed to import some documents.')
                    fail_str='\n'.join(fail_list)
                    msg.setDetailedText(fail_str)

                # only pdf annotation export failures
                elif len(fail_list)==0 and len(pdf_fail_list)>0:
                    msg.setInformativeText('Failed to export annotations in some PDFs.')
                    fail_str='\n'.join(pdf_fail_list)
                    msg.setDetailedText(fail_str)

                # both doc transfer and pdf annotation export failures
                elif len(fail_list)>0 and len(pdf_fail_list)>0:
                    msg.setInformativeText('''
                    Failed to import some documents. <br/>
                    Failed to export annotations in some PDFs.''')
                    fail_str1='\n'.join(fail_list)
                    fail_str2='\n'.join(pdf_fail_list)
                    msg.setDetailedText('Failed documents:\n%s\n\nFailed PDFs:\n%s'\
                            %(fail_str1, fail_str2))

                choice=msg.exec_()

                # open new library
                if choice==1:
                    LOGGER.info('Emitting open lib signal. File = %s' %file_name)
                    self.thread_run_dialog2.accept()
                    self.reject()
                    self.open_lib_signal.emit(file_name)

            #------------------No failed jobs------------------
            else:
                choice=QtWidgets.QMessageBox.question(self, 'Import completed',
                        'Open new library?',
                        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

                # open new library
                if choice==QtWidgets.QMessageBox.Yes:
                    LOGGER.info('Emitting open lib signal. File = %s' %file_name)
                    self.thread_run_dialog2.accept()
                    self.reject()
                    self.open_lib_signal.emit(file_name)


        #------------Failed to commit database------------
        elif rec==1:
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setWindowTitle('Oopsie')
            msg.setText("Failed to write output database file.")
            msg.exec_()

            LOGGER.warning('Failed to commit output sqlite.')
            dirname,fname=os.path.split(file_name)
            lib_folder=os.path.join(dirname,os.path.splitext(fname)[0])

            #----------------Remove sqlite file----------------
            if os.path.exists(file_name):
                os.remove(file_name)
                LOGGER.info('Remove sqlite database file %s' %file_name)
            if os.path.exists(lib_folder):
                shutil.rmtree(lib_folder)
                LOGGER.info('Remove lib folder %s' %lib_folder)

            self.thread_run_dialog2.accept()

            return

        return



    def popUpGiveName(self):

        msg=QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setWindowTitle('Input Needed')
        msg.setText("Give a name to the library.")
        msg.exec_()

        return


    def popUpGiveFile(self):

        msg=QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setWindowTitle('Input Needed')
        msg.setText("Choose file to be imported.")
        msg.exec_()

        return




'''
Import data from Mendeley.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''


import sys
import os
import sqlite3
import shutil
import re
import logging
import time
from datetime import datetime
if sys.version_info[0]>=3:
    #---------------------Python3---------------------
    from urllib.parse import unquote
    from urllib.parse import urlparse
else:
    #--------------------Python2.7--------------------
    from urllib import unquote
    from urlparse import urlparse

from . import sqlitedb
from . import exportpdf
from bs4 import BeautifulSoup


LOGGER=logging.getLogger(__name__)


# columns to read from Mendeley database
READ_DOC_ATTRS=[\
'issn', 'issue', 'language', 'read', 'type', 'confirmed',
'deduplicated', 'favourite', 'note',
'abstract', 'advisor', 'added',
'arxivId', 'title', 'pmid',
'publication', 'publicLawNumber', 'month',
'pages', 'sections', 'seriesEditor', 'series', 'seriesNumber',
'publisher', 'reprintEdition', 'reviewedArticle', 'revisionNumber',
'userType', 'volume', 'year', 'session', 'shortTitle', 'sourceType',
'code', 'codeNumber', 'codeSection', 'codeVolume', 'chapter',
'citationKey', 'city', 'day', 'department', 'doi', 'edition',
'committee', 'counsel', 'country', 'dateAccessed',
'internationalAuthor', 'internationalNumber', 'internationalTitle',
'internationalUserType', 'genre',
'institution', 'lastUpdate', 'legalStatus', 'length', 'medium', 'isbn']

# columns to create in output sqlite
WRITE_DOC_ATTRS=[\
'issn', 'issue', 'language', 'read', 'type', 'confirmed',
'deduplicated', 'favourite', 'note',
'abstract', 'advisor', 'added',
'arxivId', 'title', 'pmid',
'publication', 'publicLawNumber', 'month',
'pages', 'sections', 'seriesEditor', 'series', 'seriesNumber',
'publisher', 'reprintEdition', 'reviewedArticle', 'revisionNumber',
'userType', 'volume', 'year', 'session', 'shortTitle', 'sourceType',
'code', 'codeNumber', 'codeSection', 'codeVolume', 'chapter',
'citationKey', 'city', 'day', 'department', 'doi', 'edition',
'committee', 'counsel', 'country', 'dateAccessed',
'internationalAuthor', 'internationalNumber', 'internationalTitle',
'internationalUserType', 'genre',
'institution', 'lastUpdate', 'legalStatus', 'length', 'medium', 'isbn',
'deletionPending']

INT_COLUMNS=['month', 'year', 'day']


DOI_PATTERN=re.compile(r'(?:doi:)?\s?(10.[1-9][0-9]{3}/.*$)',
        re.DOTALL|re.UNICODE)

# Sometimes citation imported via a .bib or .ris file may contain
# a note field (`annote = {{some note}} `for .bib, `N1 - some note` for .ris).
# It can be doi strings:
#    * doi: 10.1021/ed020p517.1
#    * 10.1021/ed020p517.1
# It can also be ISBN strings: e.g. ISBN 978.....
# It can also be PMID strings: e.g. PMID: xxxx
# It could be something else, whatever the citation provider decides to put in.
# So to distinguish them from actuall notes made by users, below is a list
# of regex patterns trying to catch some recognizable patterns and exclude
# them from the notes.

NOTE_EXCLUDE_PATTERNS=[
        DOI_PATTERN,
        ]


def getDocIds(db):
    '''Get a list of docids from a folder'''

    query=\
    '''SELECT Documents.id
       FROM Documents
    '''

    ret=db.execute(query)
    data=ret.fetchall()
    docids=[ii[0] for ii in data]
    docids.sort()

    return docids


def converturl2abspath(url):
    '''Convert a url string to an absolute path
    This is necessary for filenames with unicode strings.
    '''

    #--------------------For linux--------------------
    #path = unquote(str(urlparse(url).path)).decode("utf8")
    path = unquote(str(urlparse(url).path))
    path=os.path.abspath(path)

    if os.path.exists(path):
        return path
    else:
        #-------------------For windowes-------------------
        if url[5:8]==u'///':
            url=u'file://'+url[8:]
            path=urlparse(url)
            path=os.path.join(path.netloc,path.path)
            #path=unquote(str(path)).decode('utf8')
            path=unquote(str(path))
            path=os.path.abspath(path)
            return path


def selByDocid(cursor, table, column, docid):
    """Perform SELECT query on given column(s) filtered by docid

    Args:
        cursor (sqlite connection cursor): sqlite connection cursor.
        table (str): table name.
        column (str or list/tuple): column name or names.
        docid (int): doc id.

    Returns: list of column values. If <column> is str, a list with one element.
    """

    if isinstance(column, str):
        cols=column
        single=True
    elif isinstance(column, (tuple, list)):
        cols=', '.join(column)
        single=False

    query='''SELECT %s from %s
    WHERE %s.documentId = ?''' %(cols, table, table)
    ret=cursor.execute(query, (docid,)).fetchall()

    if single:
        return [ii[0] for ii in ret]
    else:
        return [ii for ii in ret]


def convert2datetime(s):
    return datetime.strptime(s,'%Y-%m-%dT%H:%M:%SZ')


def getUserName(db):
    '''Query db to get user name'''

    query=\
    '''SELECT Profiles.firstName, Profiles.lastName
    FROM Profiles WHERE Profiles.isSelf="true"
    '''
    query_fallback=\
    '''SELECT Profiles.firstName, Profiles.lastName
    FROM Profiles
    '''
    ret=db.execute(query).fetchall()
    if len(ret)==0:
        ret=db.execute(query_fallback).fetchall()
    return ' '.join(filter(None,ret[0]))


def getFilePath(db, docid):
    """Get file path of PDF(s) using documentId

    Args:
        db (sqlite connection): connection to sqlite database.
        docid (int): doc id

    Returns: pth (list or None): None if no path is found. A list of file paths
             otherwise.
    """

    query=\
    '''SELECT Files.localUrl
       FROM Files
       LEFT JOIN DocumentFiles
           ON DocumentFiles.hash=Files.hash
       LEFT JOIN Documents
           ON Documents.id=DocumentFiles.documentId
       WHERE (Documents.id=%s)
    ''' %docid

    ret=db.execute(query)
    data=ret.fetchall()
    if len(data)==0:
        return None
    else:
        pth=[converturl2abspath(urlii[0]) for urlii in data]
        return pth


def getHighlights(db, filterdocid, results=None):
    '''Extract highlights coordinates and related meta data.

    Args:
        db (sqlite connection): connection to Mendeley sqlite database.
        filterdocid (int): doc id .
    Kwargs:
        results (dict or None): dict to store results. If None, create an
                                empty dict.

    Returns:
        results (dict): dictionary containing the query results, with the
                        following structure:

        results=
            {
            path1:
                {'highlights': {page1: [hl1, hl2,...],
                                page2: [hl1, hl2,...],
                                ...}
                }
                 'notes': {page1: [nt1, nt2,...],
                           page2: [nt1, nt2,...],
                           ...}
            path2:
                ...
            }

            where hl1={'rect': bbox,\
                       'cdate': cdate,\
                       'color': color,
                       'page':pg,
                        'author':'highlight author',\
                       'path':pth
                      }
                  note={'rect': bbox,\
                        'author':'note author',\
                        'content':docnote,\
                        'cdate': datetime.now(),\
                        'page':pg,
                        'path':pth
                        }

    '''

    # For Mendeley versions newer than 1.16.1 (include), with highlight colors
    query_new =\
    '''SELECT Files.localUrl, FileHighlightRects.page,
                    FileHighlightRects.x1, FileHighlightRects.y1,
                    FileHighlightRects.x2, FileHighlightRects.y2,
                    FileHighlights.createdTime,
                    FileHighlights.author,
                    Profiles.firstName,
                    Profiles.lastName,
                    FileHighlights.color
            FROM Files
            LEFT JOIN FileHighlights
                ON FileHighlights.fileHash=Files.hash
            LEFT JOIN FileHighlightRects
                ON FileHighlightRects.highlightId=FileHighlights.id
            LEFT JOIN Profiles
                ON Profiles.uuid=FileHighlights.profileUuid
            WHERE (FileHighlightRects.page IS NOT NULL) AND
            (FileHighlights.documentId=%s)
    ''' %filterdocid

    # For Mendeley versions older than 1.16.1, no highlight colors
    query_old =\
    '''SELECT Files.localUrl, FileHighlightRects.page,
                    FileHighlightRects.x1, FileHighlightRects.y1,
                    FileHighlightRects.x2, FileHighlightRects.y2,
                    FileHighlights.createdTime,
                    FileHighlights.author,
                    Profiles.firstName,
                    Profiles.lastName
            FROM Files
            LEFT JOIN FileHighlights
                ON FileHighlights.fileHash=Files.hash
            LEFT JOIN FileHighlightRects
                ON FileHighlightRects.highlightId=FileHighlights.id
            LEFT JOIN Profiles
                ON Profiles.uuid=FileHighlights.profileUuid
            WHERE (FileHighlightRects.page IS NOT NULL) AND
            (FileHighlights.documentId=%s)
    ''' %filterdocid

    if results is None:
        results={}

    #------------------Get highlights------------------
    try:
        ret = db.execute(query_new)
        hascolor=True
    except:
        ret = db.execute(query_old)
        hascolor=False

    for ii,r in enumerate(ret):
        pth = converturl2abspath(r[0])
        pg = r[1]
        bbox = [r[2], r[3], r[4], r[5]]
        # [x1,y1,x2,y2], (x1,y1) being bottom-left,
        # (x2,y2) being top-right. Origin at bottom-left.
        # Fix incorrect storage ordering in Mendeley:
        if bbox[0] > bbox[2]: bbox[0], bbox[2] = bbox[2], bbox[0]
        if bbox[1] > bbox[3]: bbox[1], bbox[3] = bbox[3], bbox[1]
        cdate = convert2datetime(r[6])

        # Changes suggested by matteosecli: retrieve author of highlight:
        author=r[7]
        if not author.strip():
            author=' '.join(filter(None,r[8:10]))

        color=r[10] if hascolor else None

        hlight = {'rect': bbox,\
                  'cdate': cdate,\
                  'color': color,
                  'page': pg,
                  'author': author,
                  'path': pth   # distinguish between multi-attachments
                  }

        #------------Save to dict------------
        # any better way of doing this sht?
        if pth in results:
            if 'highlights' in results[pth]:
                if pg in results[pth]['highlights']:
                    results[pth]['highlights'][pg].append(hlight)
                else:
                    results[pth]['highlights'][pg]=[hlight,]
            else:
                results[pth]['highlights']={pg:[hlight,]}
        else:
            results[pth]={'highlights': {pg:[hlight,]}}


    return results


#-------------------Get sticky notes-------------------
def getNotes(db, filterdocid, results=None):
    '''Extract notes and related meta data

    Args:
        db (sqlite connection): connection to Mendeley sqlite database.
        filterdocid (int): doc id .
    Kwargs:
        results (dict or None): dict to store results. If None, create an
                                empty dict.

    Returns:
        results (dict): dictionary containing the query results. See
                        getHighlights() for more details.
    '''

    query=\
    '''SELECT Files.localUrl, FileNotes.page,
                    FileNotes.x, FileNotes.y,
                    FileNotes.note,
                    FileNotes.modifiedTime,
                    FileNotes.author,
                    Profiles.firstName,
                    Profiles.lastName
            FROM Files
            LEFT JOIN FileNotes
                ON FileNotes.fileHash=Files.hash
            LEFT JOIN Profiles
                ON Profiles.uuid=FileNotes.profileUuid
            WHERE (FileNotes.page IS NOT NULL) AND
            (FileNotes.documentId=%s)
    ''' %filterdocid

    if results is None:
        results={}

    #------------------Get notes------------------
    ret = db.execute(query)

    for ii,r in enumerate(ret):
        pth = converturl2abspath(r[0])
        pg = r[1]
        bbox = [r[2], r[3], r[2]+30, r[3]+30]
        # needs a rectangle, size does not matter

        txt = r[4]
        cdate = convert2datetime(r[5])

        # Changes suggested by matteosecli: retrieve author of note:
        author=r[6]
        if not author.strip():
            author=' '.join(filter(None,r[7:9]))

        note = {'rect': bbox,\
                'author':author,\
                'content':txt,\
                'cdate': cdate,\
                'page':pg,
                'path':pth,
                'isgeneralnote': False
                  }

        #------------Save to dict------------
        if pth in results:
            if 'notes' in results[pth]:
                if pg in results[pth]['notes']:
                    results[pth]['notes'][pg].append(note)
                else:
                    results[pth]['notes'][pg]=[note,]
            else:
                results[pth]['notes']={pg:[note,]}
        else:
            results[pth]={'notes': {pg:[note,]}}


    return results


#-------------------Get side-bar notes-------------------
def getDocNotes(db, filterdocid, results=None):
    '''Extract side-bar notes and related meta data

    Args:
        db (sqlite connection): connection to Mendeley sqlite database.
        filterdocid (int): doc id .
    Kwargs:
        results (dict or None): dict to store results. If None, create an
                                empty dict.

    Returns:
        results (dict): dictionary containing the query results. See
                        getHighlights() for more details.
    '''

    # Some versions of Mendeley saves notes in DocumentsNotes
    query=\
    '''SELECT DocumentNotes.text,
              DocumentNotes.documentId,
              DocumentNotes.baseNote
            FROM DocumentNotes
            WHERE (DocumentNotes.documentId IS NOT NULL) AND
            (DocumentNotes.documentId=%s)
    ''' %filterdocid

    # Some versions (not sure which exactly) of Mendeley saves
    # notes in Documents.note
    query2=\
    '''SELECT Documents.note
            FROM Documents
            WHERE (Documents.note IS NOT NULL) AND
            (Documents.id=%s)
    ''' %filterdocid

    # regex to transform Mendeley's old note formatting to html
    # e.g. <m:bold>Bold</m:bold>  to <bold>Bold</bold>
    pattern=re.compile(r'<(/?)m:(bold|italic|underline|center|left|right|linebreak)(/?)>',
            re.DOTALL | re.UNICODE)
    subfunc=lambda match: u'<%s%s%s>' %match.groups()

    if results is None:
        results={}

    #------------------Get notes------------------
    ret=[]
    try:
        ret1 = db.execute(query).fetchall()
        ret.extend(ret1)
    except:
        pass
    try:
        ret2 = db.execute(query2).fetchall()
        ret.extend(ret2)
    except:
        pass
    username=getUserName(db)

    for ii,rii in enumerate(ret):
        docnote=rii[0]
        if len(docnote)==0:
            # skip u''
            continue

        # skip things that are not user notes. See def of NOTE_EXCLUDE_PATTERNS
        skip=False
        for patternii in NOTE_EXCLUDE_PATTERNS:
            if patternii.match(docnote) is not None:
                skip=True
                break

        if skip:
            continue

        docid=filterdocid
        try:
            basenote=rii[2]
        except:
            basenote=None
        pg=1

        if docnote is not None and basenote is not None\
                and docnote!=basenote:
            docnote=basenote+'\n\n'+docnote

        #--------Convert old <m:tag> to html <tag>--------
        docnote=re.sub(pattern,subfunc,docnote)

        #--------------------Parse html--------------------
        soup=BeautifulSoup(docnote,'html.parser')
        # replace <br> tags with newline
        for br in soup.find_all('br'):
            br.replace_with('\n')
        docnote=soup.get_text()
        '''
        parser=html2text.HTML2Text()
        parser.ignore_links=True
        docnote=parser.handle(docnote)
        '''

        # Try get file path
        #pth=getFilePath(db,docid) or '/pseudo_path/%s.pdf' %title
        pth=getFilePath(db,docid) # a list, could be more than 1, or None
        # If no attachment, use None as path
        if pth is None:
            # make it compatible with the for loop below
            pth=[None,]

        bbox = [50, 700, 80, 730]
        # needs a rectangle, size does not matter
        note = {'rect': bbox,
                'author': username,
                'content':docnote,
                'cdate': datetime.now(),
                'page':pg,
                'path':pth,
                'isgeneralnote': True
                  }

        #-------------------Save to dict-------------------
        # if multiple attachments, add to each of them
        for pthii in pth:
            if pthii in results:
                if 'notes' in results[pthii]:
                    if pg in results[pthii]['notes']:
                        results[pthii]['notes'][pg].insert(0,note)
                    else:
                        results[pthii]['notes'][pg]=[note,]
                else:
                    results[pthii]['notes']={pg:[note,]}
            else:
                results[pthii]={'notes': {pg:[note,]}}


    return results


def importMendeleyPreprocess(jobid, file_in_path, file_out_path):
    """Prepare for copying Mendeley data

    Args:
        jobid (int): job id.
        file_in_path (str): abspath of Mendeley sqlite database file.
        file_out_path (str): abspath of output sqlite database file.

    Returns:
        rec (int): 0 for success, 1 otherwise.
        jobid (int): the input jobd returned as it is.
        dbin (sqlite connection): connection to sqlite at <file_in_path>.
        dbout (sqlite connection): connection to sqlite at <file_out_path>.
        docids (list): list of int doc ids in Mendeley sqlite.
        lib_folder (str): path to a newly created folder to store PDF files.
        lib_name (str): name of the output library.

    In this function, an empty sqlite database is created, and folder info
    obtained from Mendeley database are copied over. A list of doc ids is
    get from Mendeley, used for data transfer in importMendeleyCopyData().
    """

    try:
        #--------------Connect input database--------------
        dbfin=os.path.abspath(file_in_path)

        try:
            dbin = sqlite3.connect(dbfin)
            LOGGER.info('Connected to database %s' %file_in_path)
        except:
            LOGGER.exception('Failed to connect to database %s' %file_in_path)
            return 1, jobid, None, None, None, None, None

        cin=dbin.cursor()

        #-----------------Create empty db-----------------
        file_out_path=os.path.expanduser(file_out_path)
        dbout, storage_folder, lib_name=sqlitedb.createNewDatabase(file_out_path)
        cout=dbout.cursor()

        lib_folder=os.path.join(storage_folder,lib_name)
        rel_lib_folder=os.path.join('', lib_name) # relative to storage folder

        LOGGER.debug('storage_folder = %s' %storage_folder)
        LOGGER.debug('lib_name = %s' %lib_name)
        LOGGER.debug('lib_folder = %s' %lib_folder)
        LOGGER.debug('rel_lib_folder = %s' %rel_lib_folder)

        #----------------Copy folders table----------------
        query='''SELECT id, name, parentId
        FROM Folders
        '''
        ret=cin.execute(query).fetchall()

        #------------If Default in folder list------------
        folders=[]
        for fid, fname, pid in ret:
            if fname=='Default' and pid==-1:
                fname='Default_Mendeley'
                LOGGER.info('Renaming "Default" folder to "Default_Mendeley')
            # may not happen to you
            if pid==0:
                pid=-1
            folders.append((fid, fname, pid))

        query='''INSERT OR IGNORE INTO Folders (id, name, parentId, path)
        VALUES (?,?,?,?)'''

        for fii in folders:
            cout.execute(query, (fii[0], fii[1], fii[2],
                os.path.join(rel_lib_folder,fii[1])))

        #----------------Loop through docs----------------
        docids=getDocIds(dbin)

        LOGGER.info('NO. of docs in database = %d' %len(docids))

        return 0, jobid, dbin, dbout, docids, lib_folder, lib_name
    except:
        return 1, jobid, None, None, None, None, None


def importMendeleyCopyData(jobid, dbin, dbout, lib_name, lib_folder,
        rename_file, ii, docii):
    """Copy document data from Mendeley and commit to output sqlite

    Args:
        jobid (int): job id.
        dbin (sqlite connection): connection to sqlite at <file_in_path>.
        dbout (sqlite connection): connection to sqlite at <file_out_path>.
        lib_folder (str): path to a newly created folder to store PDF files.
        lib_name (str): name of the output library.
        rename_file (bool): whether to rename attachment PDF files when copying.
        ii (int): index of a doc id in the doc id list. This is used as the
                  doc id in the output sqlite.
        docii (int or None): if int, id of the copied doc. If None, call
                             commit() on <dbout> and close() on <dbin>.

    Returns:
        rec (int): 0 if success copy. 1 if failed. 2 if failed to export
                   annotations in the attachment PDF(s).
        jobid (int): input jobid.
        file_fail (str): paths of PDF files that failed to get their
                         annotations exported.
    """

    if docii is None:
        # this signals db commit
        dbout.commit()
        dbout.close()
        dbin.close()
        return 0, jobid, None

    cout=dbout.cursor()
    cin=dbin.cursor()

    file_folder=os.path.join(lib_folder,'_collections')
    rel_lib_folder=os.path.join('', lib_name) # relative to storage folder
    rel_file_folder=os.path.join('','_collections')

    try:
        ii+=1

        LOGGER.debug('Copying doc id = %s' %docii)

        #---------------Get Documents columns---------------
        meta_query='''SELECT %s from Documents
        WHERE Documents.id = ?''' %', '.join(READ_DOC_ATTRS)

        ret=cin.execute(meta_query, (docii,))
        metaii=list(ret.fetchall()[0])

        # convert int
        for jj in INT_COLUMNS:
            idxjj=READ_DOC_ATTRS.index(jj)
            fjj=metaii[idxjj]
            if fjj is not None:
                fjj=int(fjj)
                metaii[idxjj]=fjj

        # make sure added exists
        idxjj=READ_DOC_ATTRS.index('added')
        fjj=metaii[idxjj]
        if fjj is None:
            fjj=str(int(time.time()))
        metaii[idxjj]=fjj

        meta_dictii=dict(zip(READ_DOC_ATTRS, metaii))
        meta_dictii['deletionPending']='false'
        metaii.append('false')
        metaii=tuple(metaii)

        #-----------------Get DocumentTags-----------------
        tags=selByDocid(cin, 'DocumentTags', 'tag', docii)

        #------------------Get FileNotes------------------
        #notes=selByDocid(cin, 'DocumentNotes', 'text', docii)
        notes=selByDocid(cin, 'FileNotes', ['note', 'modifiedTime', 'createdTime'], docii)

        #---------------Get DocumentKeywords---------------
        keywords=selByDocid(cin, 'DocumentKeywords', 'keyword', docii)

        #-----------------Get DocumentUrls-----------------
        urls=selByDocid(cin, 'DocumentUrls', 'url', docii)
        urls=[str(bjj) for bjj in urls]   # convert blob to str

        #---------------Get DocumentFolders---------------
        query='''SELECT Folders.id, Folders.name, Folders.parentId
        FROM Folders
        LEFT JOIN DocumentFolders ON DocumentFolders.folderId = Folders.id
        WHERE DocumentFolders.documentId = ?'''

        ret=cin.execute(query, (docii,)).fetchall()

        # if name conflict with Default
        folder_info=[]
        for fid, fname, pid in ret:
            if fname=='Default' and pid==-1:
                fname='Default_Mendeley'
            if pid==0:
                pid=-1
            folder_info.append((fid, fname, pid))

        # if not in any folder, put to Default
        if len(folder_info)==0:
            folder_info.append((0, 'Default', -1))

        #-------------Get DocumentContributors-------------
        authors=selByDocid(cin, 'DocumentContributors',
        ['contribution', 'firstNames', 'Lastname'], docii)

        meta_dictii['lastName_l']=[jj[2] for jj in authors]

        #------------Insert to output database------------
        query='''INSERT INTO Documents (
        %s )
        VALUES (%s)''' %(', '.join(WRITE_DOC_ATTRS), ', '.join(['?']*len(WRITE_DOC_ATTRS)))

        metaii=tuple([meta_dictii[jj] for jj in WRITE_DOC_ATTRS])
        cout.execute(query, metaii)

        for tagii in tags:
            query='''INSERT INTO DocumentTags (did, tag)
            VALUES (?, ?)'''
            cout.execute(query, (ii, tagii))

        for keyii in keywords:
            query='''INSERT INTO DocumentKeywords (did, text)
            VALUES (?, ?)'''
            cout.execute(query, (ii, keyii))

        for nii in notes:
            query='''INSERT INTO DocumentNotes (did, note, modifiedTime, createdTime)
            VALUES (?, ?, ?, ?)'''
            cout.execute(query, (ii,)+ nii)

        for fii in folder_info:
            query='''INSERT INTO DocumentFolders (did, folderid)
            VALUES (?, ?)'''
            cout.execute(query, (ii, fii[0]))

            query='''INSERT OR IGNORE INTO Folders (id, name, parentId, path)
            VALUES (?,?,?,?)'''
            cout.execute(query, (fii[0], fii[1], fii[2],
                os.path.join(rel_lib_folder,fii[1])))

        for aii in authors:
            query='''INSERT INTO DocumentContributors (
            did, contribution, firstNames, lastName)
            VALUES (?, ?, ?, ?)'''
            cout.execute(query, (ii,)+aii)

        for urlii in urls:
            query='''INSERT INTO DocumentUrls (did, url)
            VALUES (?, ?)'''
            cout.execute(query, (ii, urlii))

        #---------------Fetch pdf highlights and notes---------
        annotations={}
        annotations = getHighlights(dbin,docii,annotations)
        annotations = getNotes(dbin,docii,annotations)
        annotations = getDocNotes(dbin,docii,annotations)

        #------------------Get file paths------------------
        query='''SELECT Files.localUrl
        FROM Files
        LEFT JOIN DocumentFiles ON DocumentFiles.hash = Files.hash
        WHERE DocumentFiles.documentId = ?'''

        ret=cin.execute(query, (docii,))

        fileurl=ret.fetchall()
        meta_dictii['files_l']=[]

        file_fail_list=[]
        if len(fileurl)>0:

            for fileii in fileurl:
                urlii=fileii[0]
                filepath=converturl2abspath(urlii)
                meta_dictii['files_l'].append(filepath)

            for filepath in meta_dictii['files_l']:

                #-------------------Rename file-------------------
                if rename_file:
                    filename=sqlitedb.renameFile(filepath,meta_dictii)
                else:
                    filename=os.path.split(filepath)[1]

                #---------------Remove invalid chars---------------
                filename=re.sub(r'[//\ <>:"|?*]','_',filename)
                filename=re.sub(r'al.','al',filename)
                filename=re.sub(r'_-_','_',filename)
                filename=filename.strip()

                relpath=os.path.join(rel_file_folder,filename)

                query='''INSERT INTO DocumentFiles (did, relpath)
                VALUES (?, ?)'''
                cout.execute(query, (ii, relpath))

                #---------------------Copy file---------------------
                abspath=os.path.join(file_folder, filename)

                #LOGGER.debug('relpath = %s' %relpath)
                LOGGER.debug('abspath = %s' %abspath)

                if len(annotations)>0 and filepath in annotations:
                    anno=annotations[filepath]
                    try:
                        exportpdf.exportPdf(filepath, abspath, anno)
                    except:
                        LOGGER.warning('Failed to export annotated pdf %s'\
                                %filepath)
                        file_fail_list.append(filepath)
                        try:
                            shutil.copy2(filepath,abspath)
                            LOGGER.debug('Copied %s to %s' %(filepath, abspath))
                        except:
                            LOGGER.exception('Failed to copy %s to %s' %(filepath,
                                abspath))
                else:
                    try:
                        shutil.copy2(filepath,abspath)
                        LOGGER.debug('Copied %s to %s' %(filepath, abspath))
                    except:
                        LOGGER.exception('Failed to copy %s to %s' %(filepath,
                            abspath))

        if len(file_fail_list)>0:
            return 2, jobid, '; '.join(file_fail_list)
        else:
            return 0, jobid, None
    except:
        return 1, jobid, None




def importMendeley(file_in_path, file_out_path, rename_file):
    '''For debug'''

    #--------------Connect input database--------------
    dbfin=os.path.abspath(file_in_path)

    try:
        dbin = sqlite3.connect(dbfin)
        print('Connected to database:')
    except:
        print('Failed to connect to database:')

    cin=dbin.cursor()

    #-----------------Create empty db-----------------
    file_out_path=os.path.expanduser(file_out_path)
    dbout, storage_folder, lib_name=sqlitedb.createNewDatabase(file_out_path)
    cout=dbout.cursor()

    lib_folder=os.path.join(storage_folder,lib_name)
    file_folder=os.path.join(lib_folder,'_collections')
    rel_lib_folder=os.path.join('', lib_name) # relative to storage folder
    rel_file_folder=os.path.join('','_collections')

    #----------------Copy folders table----------------
    query='''SELECT id, name, parentId
    FROM Folders
    '''

    ret=cin.execute(query).fetchall()

    #------------If Default in folder list------------
    folders=[]
    for fid, fname, pid in ret:
        if fname=='Default' and pid==-1:
            fname='Default_Mendeley'
        # may not happen to you
        if pid==0:
            pid=-1
        folders.append((fid, fname, pid))

    query='''INSERT OR IGNORE INTO Folders (id, name, parentId, path)
    VALUES (?,?,?,?)'''

    for fii in folders:
        cout.execute(query, (fii[0], fii[1], fii[2],
            os.path.join(rel_lib_folder,fii[1])))

    #----------------Loop through docs----------------
    docids=getDocIds(dbin)

    for ii,docii in enumerate(docids):

        ii+=1
        print('# <imporMendeley>: Copying doc ', ii)

        #---------------Get Documents columns---------------
        meta_query='''SELECT %s from Documents
        WHERE Documents.id = ?''' %', '.join(READ_DOC_ATTRS)

        ret=cin.execute(meta_query, (docii,))
        metaii=list(ret.fetchall()[0])

        # convert int
        for jj in INT_COLUMNS:
            idxjj=READ_DOC_ATTRS.index(jj)
            fjj=metaii[idxjj]
            if fjj is not None:
                fjj=int(fjj)
                metaii[idxjj]=fjj

        # make sure added exists
        idxjj=READ_DOC_ATTRS.index('added')
        fjj=metaii[idxjj]
        if fjj is None:
            fjj=str(int(time.time()))
        metaii[idxjj]=fjj

        meta_dictii=dict(zip(READ_DOC_ATTRS, metaii))
        meta_dictii['deletionPending']='false'
        metaii.append('false')
        metaii=tuple(metaii)

        #-----------------Get DocumentTags-----------------
        tags=selByDocid(cin, 'DocumentTags', 'tag', docii)

        #------------------Get FileNotes------------------
        #notes=selByDocid(cin, 'DocumentNotes', 'text', docii)
        notes=selByDocid(cin, 'FileNotes', ['note', 'modifiedTime', 'createdTime'], docii)

        #---------------Get DocumentKeywords---------------
        keywords=selByDocid(cin, 'DocumentKeywords', 'keyword', docii)

        #-----------------Get DocumentUrls-----------------
        urls=selByDocid(cin, 'DocumentUrls', 'url', docii)
        urls=[str(bjj) for bjj in urls]   # convert blob to str

        #---------------Get DocumentFolders---------------
        query='''SELECT Folders.id, Folders.name, Folders.parentId
        FROM Folders
        LEFT JOIN DocumentFolders ON DocumentFolders.folderId = Folders.id
        WHERE DocumentFolders.documentId = ?'''

        ret=cin.execute(query, (docii,)).fetchall()

        # if name conflict with Default
        folder_info=[]
        for fid, fname, pid in ret:
            if fname=='Default' and pid==-1:
                fname='Default_Mendeley'
            if pid==0:
                pid=-1
            folder_info.append((fid, fname, pid))

        # if not in any folder, put to Default
        if len(folder_info)==0:
            folder_info.append((0, 'Default', -1))

        #-------------Get DocumentContributors-------------
        authors=selByDocid(cin, 'DocumentContributors',
        ['contribution', 'firstNames', 'Lastname'], docii)

        meta_dictii['lastName_l']=[jj[2] for jj in authors]

        #------------Insert to output database------------
        query='''INSERT INTO Documents (
        %s )
        VALUES (%s)''' %(', '.join(WRITE_DOC_ATTRS), ', '.join(['?']*len(WRITE_DOC_ATTRS)))

        metaii=tuple([meta_dictii[jj] for jj in WRITE_DOC_ATTRS])
        cout.execute(query, metaii)

        for tagii in tags:
            query='''INSERT INTO DocumentTags (did, tag)
            VALUES (?, ?)'''
            cout.execute(query, (ii, tagii))

        for keyii in keywords:
            query='''INSERT INTO DocumentKeywords (did, text)
            VALUES (?, ?)'''
            cout.execute(query, (ii, keyii))

        for nii in notes:
            query='''INSERT INTO DocumentNotes (did, note, modifiedTime, createdTime)
            VALUES (?, ?, ?, ?)'''
            cout.execute(query, (ii,)+ nii)

        for fii in folder_info:
            query='''INSERT INTO DocumentFolders (did, folderid)
            VALUES (?, ?)'''
            cout.execute(query, (ii, fii[0]))

            query='''INSERT OR IGNORE INTO Folders (id, name, parentId, path)
            VALUES (?,?,?,?)'''
            cout.execute(query, (fii[0], fii[1], fii[2],
                os.path.join(rel_lib_folder,fii[1])))

        for aii in authors:
            query='''INSERT INTO DocumentContributors (
            did, contribution, firstNames, lastName)
            VALUES (?, ?, ?, ?)'''
            cout.execute(query, (ii,)+aii)

        for urlii in urls:
            query='''INSERT INTO DocumentUrls (did, url)
            VALUES (?, ?)'''
            cout.execute(query, (ii, urlii))

        # get path
        query='''SELECT Files.localUrl
        FROM Files
        LEFT JOIN DocumentFiles ON DocumentFiles.hash = Files.hash
        WHERE DocumentFiles.documentId = ?'''

        ret=cin.execute(query, (docii,))

        fileurl=ret.fetchall()
        meta_dictii['files_l']=[]
        if len(fileurl)>0:

            for fileii in fileurl:
                urlii=fileii[0]
                filepath=converturl2abspath(urlii)
                meta_dictii['files_l'].append(filepath)

            for filepath in meta_dictii['files_l']:

                # rename file
                if rename_file:
                    filename=sqlitedb.renameFile(filepath,meta_dictii)
                else:
                    filename=os.path.split(filepath)[1]

                filename=re.sub(r'[//\ <>:"|?*]','_',filename)
                filename=re.sub(r'al.','al',filename)
                filename=re.sub(r'_-_','_',filename)
                filename=filename.strip()

                relpath=os.path.join(rel_file_folder,filename)

                query='''INSERT INTO DocumentFiles (did, relpath)
                VALUES (?, ?)'''

                cout.execute(query, (ii, relpath))

                abspath=os.path.join(file_folder, filename)
                shutil.copy2(filepath,abspath)

    dbout.commit()

    return 0


if __name__=='__main__':

    #importMendeley(FILE_IN_NAME, FILE_OUT_NAME, True)
    FILE_OUT_NAME='../New_Folder/men.sqlite'
    FILE_IN_NAME='../mendeley.sqlite'
    rec, jobid, dbin, dbout, docids, lib_folder, lib_name=\
    importMendeleyPreprocess(0, FILE_IN_NAME, FILE_OUT_NAME)

    for ii, docii in enumerate(docids):
        rec2, jobid, fail_list = importMendeleyCopyData(1, dbin, dbout, lib_name, lib_folder,
            True, ii, docii)

        print('rec = ', rec2, 'jobid = ', jobid, 'fail=', fail_list)



'''
In-memory meta data dict (DocMeta) definition.
Sqlite database read and write functions.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import shutil
import time
import re
from datetime import datetime
import sqlite3
import logging
from collections import MutableMapping
try:
    from . import bibparse
except:
    import bibparse
try:
    from .tools import autoRename
except:
    from tools import autoRename

DOC_ATTRS=[\
'issn', 'issue', 'language', 'read', 'type', 'confirmed',
'deduplicated', 'favourite', 'note',
'abstract', 'advisor', 'added',
'arxivId', 'title', 'pmid',
'publication', 'publicLawNumber', 'month',
'pages', 'sections', 'seriesEditor', 'series', 'seriesNumber',
'publisher', 'reprintEdition', 'reviewedArticle', 'revisionNumber',
'userType', 'volume', 'year', 'session', 'shortTitle', 'sourceType',
'code', 'codeNumber', 'codeSection', 'codeVolume', 'chapter',
'citationKey', 'city', 'day', 'department', 'doi', 'edition',
'committee', 'counsel', 'country', 'dateAccessed',
'internationalAuthor', 'internationalNumber', 'internationalTitle',
'internationalUserType', 'genre',
'institution', 'lastUpdate', 'legalStatus', 'length', 'medium', 'isbn',
'deletionPending']

INT_COLUMNS=['month', 'year', 'day']

LOGGER=logging.getLogger(__name__)


class DocMeta(MutableMapping):
    '''A custom dict definition storing meta data of a document

    Allow only a given list of keys and restrict keys to str type.
    Give default values to keys.
    Some values are returned in a getter manner.

    keys ending with '_l' suffix denotes a list value, e.g. 'authors_l'.
    '''

    def __init__(self, *args, **kwargs):
        # set defaults
        self.store = {
                'id': None, 'title': None, 'issue': None, 'pages': None,
                'publication': None, 'volume': None, 'year': None,
                'doi': None, 'abstract': None, 'arxivId': None, 'chapter': None,
                'city': None, 'country': None, 'edition': None,
                'institution': None, 'isbn': None, 'issn': None, 'month': None,
                'day': None, 'publisher': None, 'series': None,
                'type': 'article',
                'read': 'false', 'favourite': 'false',
                'pmid': None, 'added': str(int(time.time())),
                'confirmed': 'false',
                'firstNames_l': [],
                'lastName_l': [],
                'keywords_l': [],
                'files_l': [],
                'folders_l': [],
                'tags_l': [],
                'urls_l': [],
                'deletionPending': 'false',
                'notes': None
                }

        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        if key == 'has_file':
            return True if len(self.store['files_l']) > 0 else False
        elif key == 'authors_l':
            return zipAuthors(self.store['firstNames_l'],
                    self.store['lastName_l'])
        elif key == 'citationkey':
            ck = self.store.get('citationkey',None)
            if ck:
                return ck
            else:
                last = self.store.get('lastName_l',[])
                year = self.store.get('year',None)
                if len(last) > 0 and year:
                    ck='%s%s' %(last[0],str(year))
                    return ck
                else:
                    return ''
        else:
            return self.store[key]

    def __setitem__(self, key, value):
        if not isinstance(key,str):
            raise Exception("accept only str type keys. key: %s, value: %s"\
                    %(key, value))
        if key.endswith('_l'):
            if not isinstance(value,(tuple,list)):
                raise Exception("keys end with '_l' accepts only list or tuple. key: %s, value: %s" %(key, value))
        if key not in self.store.keys():
            return
        if value is None:
            return

        self.store[key] = value

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __repr__(self):
        return self.store.__repr__()



def readSqlite(dbin):
    """Read sqlite data

    Args:
        dbin (sqlite connection): connection to sqlite.

    Returns:
        meta (dict): meta data of all documents. keys: docid,
            values: DocMeta dict.
        folder_data (dict): documents in each folder. keys: folder id in str,
            values: list of doc ids.
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
    """

    #-------------------Get folders-------------------
    folder_dict=getFolders(dbin)
    LOGGER.debug('Got %d folders from database.' %len(folder_dict))

    #-------------------Get metadata-------------------
    meta={}

    # this is when I tried fts5 tables in sqlite where 'rowid' was hardwared
    # by sqlite, and eventually switched back to normal table.
    cursor=dbin.execute('SELECT * FROM Documents')
    names=list(map(lambda x:x[0], cursor.description))
    query='''SELECT DISTINCT %s
    FROM Documents
    ''' %('id' if 'id' in names else 'rowid')

    docids=dbin.execute(query).fetchall()
    docids=[ii[0] for ii in docids]
    docids.sort()

    folder_data={}
    folder_data['-2']=[] # needs review folder
    folder_data['-3']=[] # trash can folder

    for idii in docids:

        metaii=getMetaData(dbin,idii)
        meta[idii]=metaii

        if metaii['confirmed'] is None or metaii['confirmed']=='false':
            folder_data['-2'].append(idii)
        if metaii['deletionPending']=='true' and len(metaii['folders_l'])==0:
            folder_data['-3'].append(idii)

        folderii=metaii['folders_l'] # (folderid, foldername)

        # convert folder id to str, as QListWidgetItem request str. very annoying
        # remember to convert back to int when writing to sqlite
        folderids=[str(ff[0]) for ff in folderii]
        for fii in folderids:
            if fii in folder_data:
                folder_data[fii].append(idii)
            else:
                folder_data[fii]=[idii]

    #----------------Add empty folders----------------
    empty_folderids=list(set(folder_dict.keys()).difference(folder_data.keys()))
    for fii in empty_folderids:
        folder_data[fii]=[]

    LOGGER.info('Done reading in sqlite database.')
    LOGGER.info('len(meta) = %d. len(folder_dict) = %d'\
            %(len(meta), len(folder_dict)))

    return meta, folder_data, folder_dict


def fetchField(db, query, values, ncol=1, ret_type='str'):
    """Query columns from sqlite and do some formatting

    Args:
        db (sqlite connection): sqlite connection.
        query (str): SELECT statement template.
        values (tuple): value list used in sqlite execute.

    Kwargs:
        ncol (int): number of columns queried in <query>.
        ret_type (str): if 'str' and ncol==1, return a single string.
                        if 'str' and ncol>1, return a list of strings.
                        if 'list', return a list.

    Returns: aa: column values. See ret_type.
    """

    if ret_type not in ['str','list']:
        raise Exception("<ret_type> is one of ['str','list'].")

    aa=db.execute(query,values).fetchall()

    if len(aa)==0:
        # don't think this ever happens. Null is returned as None
        if ret_type=='str':
            return None
        else:
            return []

    if ncol==1:
        aa=[ii[0] for ii in aa]
    if ret_type=='str':
        if len(aa)==1:
            if aa[0] is None:
                return None
            else:
                return str(aa[0])
        else:
            return '; '.join(aa)
    else:
        return aa


def getMetaData(db, did):
    """Get meta data of a given document from sqlite

    Args:
        db (sqlite connection): sqlite connection.
        did (int): id of doc.

    Returns: result (DocMeta): meta data dict.
    """

    cursor=db.execute('SELECT * FROM Documents')
    names=list(map(lambda x:x[0], cursor.description))

    # fetch column from Document table
    if 'id' in names:
        query_base=\
        '''SELECT Documents.%s
           FROM Documents
           WHERE (Documents.id=?)
        '''
    else:
        query_base=\
        '''SELECT Documents.%s
           FROM Documents
           WHERE (Documents.rowid=?)
        '''

    query_tags=\
    '''
    SELECT DocumentTags.tag
    FROM DocumentTags
    WHERE (DocumentTags.did=?)
    '''

    query_urls=\
    '''
    SELECT DocumentUrls.url
    FROM DocumentUrls
    WHERE (DocumentUrls.did=?)
    '''

    query_firstnames=\
    '''
    SELECT DocumentContributors.firstNames
    FROM DocumentContributors
    WHERE (DocumentContributors.did=?)
    '''

    query_lastnames=\
    '''
    SELECT DocumentContributors.lastName
    FROM DocumentContributors
    WHERE (DocumentContributors.did=?)
    '''

    query_keywords=\
    '''
    SELECT DocumentKeywords.text
    FROM DocumentKeywords
    WHERE (DocumentKeywords.did=?)
    '''

    query_folder=\
    '''
    SELECT Folders.id, Folders.name
    FROM Folders
    LEFT JOIN DocumentFolders ON DocumentFolders.folderid=Folders.id
    WHERE (DocumentFolders.did=?)
    '''

    query_files=\
    '''
    SELECT DocumentFiles.relpath
    FROM DocumentFiles
    WHERE (DocumentFiles.did=?)
    '''

    query_notes=\
    '''
    SELECT DocumentNotes.note
    FROM DocumentNotes
    WHERE (DocumentNotes.did=?)
    '''

    #------------------Get file meta data------------------
    fields=['citationkey','title','issue','pages',\
            'publication','volume','year','doi','abstract',\
            'arxivId','chapter','city','country','edition','institution',\
            'isbn','issn','month','day','publisher','series','type',\
            'read','favourite','pmid','added','confirmed', 'deletionPending']

    result=DocMeta()

    if 'id' in names:
        vii=fetchField(db,query_base %'id', (did,))
        result['id']=int(vii)
    else:
        vii=fetchField(db,query_base %'rowid', (did,))
        result['rowid']=vii

    # query single-worded fields, e.g. year, city
    for kii in fields:
        vii=fetchField(db,query_base %(kii), (did,))
        result[kii]=vii

    # query notes
    result['notes']=fetchField(db,query_notes,(did,),1,'str')

    # query list fields, .e.g firstnames, tags
    result['firstNames_l']=fetchField(db,query_firstnames,(did,),1,'list')
    result['lastName_l']=fetchField(db,query_lastnames,(did,),1,'list')
    result['keywords_l']=fetchField(db,query_keywords,(did,),1,'list')
    result['files_l']=fetchField(db,query_files,(did,),1,'list')
    result['folders_l']=fetchField(db,query_folder,(did,),2,'list')
    result['tags_l']=fetchField(db,query_tags,(did,),1,'list')
    result['urls_l']=fetchField(db,query_urls,(did,),1,'list')

    #LOGGER.debug('Done fetching meta data for doc %s' %str(did))

    return result


def zipAuthors(firstnames, lastnames):
    """Create author name list from lists of first names and last names.

    Args:
        firstnames (list): list of first names.
        lastnames (list): list of last names.

    Returns: authors (list): author names list.

    """

    if len(firstnames)!=len(lastnames):
        LOGGER.error('Length of firstname and lastname dont match.')
        LOGGER.error('firstnames = %s' %firstnames)
        LOGGER.error('lastnames = %s' %lastnames)
        raise Exception("Length of firstname and lastname dont match.")
    authors=[]
    for ii in range(len(firstnames)):
        fii=firstnames[ii]
        lii=lastnames[ii]
        if fii!='' and lii!='':
            authors.append('%s, %s' %(lii.strip(),fii.strip()))
        elif fii=='' and lii!='':
            authors.append(lii)
        elif fii!='' and lii=='':
            authors.append(fii)

    return authors


def getSubFolders(folder_dict, folderid):
    """Get all subfolders of a give folder

    Args:
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
        folderid (str): id of given folder.

    Returns: results (list): list of folder ids. Subfolders walked from given
             folder.
    """

    results=[kk for kk,vv in folder_dict.items() if vv[1]==folderid]

    subs=[]
    for fii in results:
        subids=getSubFolders(folder_dict,fii)
        subs.extend(subids)

    results.extend(subs)

    return results


def getTrashedFolders(folder_dict):
    """Get all folders inside Trash

    Args:
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.

    Returns: results (list): list of folder ids. Subfolders walked from the
             Trash (id='-3') folder.
    """
    '''
    results=[kk for kk,vv in folder_dict.items() if vv[1]=='-3']


    def _getChildFolder(folder_dict,folderid,results=None):
        if results is None:
            results=[]
        for idii in folder_dict:
            fii,pii=folder_dict[idii]
            if pii==folderid:
                results.append(idii)
                results=_getChildFolder(folder_dict,fii,results)
        return results
    for fii in results:
        subids=_getChildFolder(folder_dict,fii)
        results.extend(subids)

    return results
    '''
    results = getSubFolders(folder_dict, '-3')
    LOGGER.debug('Ids of trashed folders = %s' %results)
    return results


def walkFolderTree(folder_dict, folder_data, folderid, docids=None,
        folderids=None):
    """Get subfolders walked from a given folder and all docs within

    Args:
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
        folder_data (dict): documents in each folder. keys: folder id in str,
            values: list of doc ids.
        folderid (str): id of folder to start the walk.

    Kwargs:
        docids (list): list storing the doc ids in folders visited during the
                       walk. If None, create an empty list.
        folderids (list): list storing the ids of folders visited during the
                       walk. If None, create an empty list.

    Returns:
        docids (list): list storing the doc ids in folders visited during the
                       walk.
        folderids (list): list storing the ids of folders visited during the
                       walk.

    """

    if docids is None:
        docids=[]
    if folderids is None:
        folderids=[]

    docids.extend(folder_data[folderid])
    folderids.append(folderid)

    subfolderids=getChildFolders(folder_dict,folderid)
    for sii in subfolderids:
        folderids,docids=walkFolderTree(folder_dict,folder_data,sii,
                docids,folderids)

    folderids=list(set(folderids))
    docids=list(set(docids))

    return folderids,docids


def findOrphanDocs(folder_data, docids, trashed_folder_ids):
    """Identify orphan docs from a given list of doc ids

    Args:
        folder_data (dict): documents in each folder. keys: folder id in str,
            values: list of doc ids.
        docids (list): list of doc ids, within which orphan docs are searched.
        trashed_folder_ids (list): list of ids of folders inside Trash.

    Returns: result (list): list of orphan doc ids.

    Orphan docs are defined as those that ONLY appear in Trash, or folder(s)
    inside Trash.
    """

    idsinfolders=[]

    for kk,vv in folder_data.items():
        if kk not in ['-1','-2','-3']+trashed_folder_ids:
            idsinfolders.extend(vv)

    result=[idii for idii in docids if idii not in idsinfolders]

    LOGGER.debug('excluded folders = ["-1","-2","-3"]+%s' %trashed_folder_ids)
    LOGGER.debug('Orphan docs = %s' %result)

    return result


def findOrphanDocs2(db):
    '''Find orphan docs from sqlite data

    Not in use.
    '''

    cin=db.cursor()
    query='''SELECT rowid FROM Documents
    WHERE Documents.deletionPending='true'
    '''
    ret=cin.execute(query)
    ret=[ii[0] for ii in ret]

    LOGGER.debug('Orphan docs = %s' %ret)

    return ret


def getFolders(db):
    """Read folder info from sqlite

    Args:
        db (sqlite connection): sqlite connection.

    Returns: df (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
    """

    #-----------------Get all folders-----------------
    query='''SELECT id, name, parentId
    FROM Folders
    '''
    ret=db.execute(query)
    data=ret.fetchall()

    # dict, key: folderid, value: (folder_name, parent_id)
    # note: convert id to str
    df=dict([(str(ii[0]), (ii[1], str(ii[2]))) for ii in data])

    return df


def fetchMetaData(meta_dict, key, docids, unique, sort):
    """Get the meta data values of a given key within a list of docs

    Args:
        meta_dict (dict): meta data of all documents. keys: docid,
            values: DocMeta dict.
        key (str): key in DocMeta to get value.
        docids (list): list of doc ids within which search is done.
        unique (bool): remove duplicate or not.
        sort (bool): sort result list or not.

    Returns: result (list): list of values.

    """
    if not isinstance(docids, (tuple,list)):
        docids=[docids,]

    result=[]
    for idii in docids:
        vv=meta_dict[idii].get(key,None)
        # NOTE: don't use if vv:
        # as there are '' entries that will also trigger the if
        if vv is not None:
            if isinstance(vv, (tuple,list)):
                result.extend(vv)
            else:
                result.append(vv)

    if unique:
        result=list(set(result))
    if sort:
        result.sort()

    return result


def filterDocs(meta_dict, folder_data, filter_type, filter_text,
        current_folder):
    """Filter docs using a given filter text

    Args:
        meta_dict (dict): meta data of all documents. keys: docid,
            values: DocMeta dict.
        folder_data (dict): documents in each folder. keys: folder id in str,
            values: list of doc ids.
        filter_type (str): defines the field of the filter_text.
        filter_text (str): filtering text.
        current_folder (str): id of folder. Docs are selected from docs in
                              this folder.

    Returns: results (list): ids of docs within folder given by
             <current_folder>, containing the text <filter_text> in the field
             defined by <filter_type>.
    """

    results=[]
    if current_folder=='-1':
        docids=meta_dict.keys()
    else:
        docids=folder_data[current_folder]

    if filter_type=='Filter by authors':
        t_last,t_first=map(str.strip,filter_text.split(','))
        for kk in docids:
            authors=meta_dict[kk]['authors_l']
            if filter_text in authors:
                results.append(kk)

    elif filter_type=='Filter by tags':
        for kk in docids:
            tags=meta_dict[kk]['tags_l'] or []
            if filter_text in tags:
                results.append(kk)

    elif filter_type=='Filter by publications':
        for kk in docids:
            pubs=meta_dict[kk]['publication'] or []
            if filter_text in pubs:
                results.append(kk)

    elif filter_type=='Filter by keywords':
        for kk in docids:
            keywords=meta_dict[kk]['keywords_l'] or []
            if filter_text in keywords:
                results.append(kk)


    return results


def getChildFolders(folder_dict, folderid):
    """Get child folders of a given folder

    Args:
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
        folderid (str): id of given folder.

    Returns: results (list): list of ids of child folders.

    NOTE: different from getSubFolders() that this only get direct child
    folders of a given folder, while getSubFolders() walks done the folder
    tree.
    """

    results=[]
    for idii in folder_dict:
        fii,pii=folder_dict[idii]
        if pii==folderid:
            results.append(idii)
    results.sort()
    return results


def getFolderTree(folder_dict, folderid):
    """Get a str representation of a given folder in the folder tree

    Args:
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
        folderid (str): id of given folder.

    Returns:
        folderid (str): input folder id.
        folder (str): path of the given folder in the tree.
    """

    getFolderName=lambda folder_dict,id: folder_dict[id][0]
    getParentId=lambda folder_dict,id: folder_dict[id][1]

    folder=getFolderName(folder_dict,folderid)

    #------------Back track tree structure------------
    cid=folderid
    while True:
        pid=str(getParentId(folder_dict,cid))
        if pid=='-1':
            break
        else:
            pfolder=getFolderName(folder_dict,pid)
            folder=os.path.join(pfolder,folder)
        cid=pid

    return folderid,folder


def renameFile(fname, meta_dict, replace_space=False):
    """Rename a attachment file using meta data

    Args:
        fname (str): abspath of a file.
        meta_dict (DocMeta): meta data dict.

    Kwargs:
        replace_space (bool): whether to replace spaces in file name with some
                              symbol.

    Returns: fname2 (str): renamed file name. NOTE this doesn't include folder
                           path, ONLY file name.

    The file name is constrcuted from meta data in the format:

        Author_year_title.ext

    If the renamed abspath is too long (>255-6), the title field is cropped.
    If a doc contains more than 1 attachment files, an int index is appended
    before extension to distinguish them.
    Chars that are invalid in Linux, Mac or Win are removed, such as ':', '?'.
    """

    dirjj,filenamejj=os.path.split(fname)
    basename,ext=os.path.splitext(filenamejj)

    # get first author
    author=meta_dict['lastName_l']
    if len(author)>0:
        author=author[0]
        if author is None or author=='':
            author='Unknown'
    else:
        author='Unknown'

    # get year
    year=meta_dict['year']
    if year is None:
        year='unknown'

    # get title
    title=meta_dict.get('title',basename)

    LOGGER.debug('author = %s, year = %s, title = %s' %(author,year,title))

    # crop length
    fname2='%s_%s_%s%s' %(author,year,title,ext)
    len1=len(dirjj)+1+len(fname2)
    if len1>255-6:
        title=title[:255-6-len1]

        LOGGER.debug('Cropped title = %s' %title)

    #---------Handle multiple files for a doc---------
    if len(meta_dict['files_l'])==1:
        fname2='%s_%s_%s%s' %(author,year,title,ext)
    else:
        fname2='%s_%s_%s_%d%s' %(author,year,title,\
                meta_dict['files_l'].index(fname), ext)

        LOGGER.debug('Appending idx to a mutli-file doc.')

    if replace_space:
        fname2=fname2.replace(' ','-')

    fname2=re.sub(r'[<>:"|?*]','_',fname2)

    LOGGER.info('Old filename = %s. New filename = %s' %(fname,fname2))

    return fname2


def saveFoldersToDatabase(db, folder_ids, folder_dict, folder_data,
        lib_folder):
    """Save folder changes to sqlite

    Args:
        db (sqlite connection): connection to sqlite.
        folder_ids (list): list of ids (in str) of folders to save.
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
        folder_data (dict): documents in each folder. keys: folder id in str,
            values: list of doc ids.
        lib_folder (str): abspath to the folder of the library. By design
                          this should point to the folder CONTAINING the
                          sqlite database file.

    Returns: 0
    """

    cout=db.cursor()
    folder_ids=list(set(folder_ids))
    LOGGER.info('Saving folders with ids = %s' %folder_ids)

    for idii in folder_ids:

        folder=folder_dict.get(idii)
        if folder is None:
            # deleting folder
            LOGGER.info('Deleting folder with id = %s' %idii)
            query='''
            DELETE FROM Folders WHERE Folders.id=?
            '''
            cout.execute(query, (int(idii),))
        else:
            nameii,pidii=folder_dict[idii]
            idii=int(idii)
            pathii=os.path.join(lib_folder,nameii)
            pathii=os.path.abspath(pathii)

            LOGGER.info('Updating folder, id = %s, name = %s, pid = %s, path = %s'\
                    %(idii, nameii, pidii, pathii))

            query='''INSERT OR REPLACE INTO Folders (id, name, parentId, path) \
                     VALUES (?,?,?,?)'''
            cout.execute(query, (idii, nameii, pidii, pathii))

    db.commit()
    LOGGER.info('Done saving folders')

    return 0


def createNewDatabase(file_path):
    """Create a new sqlite database for a new library

    Args:
        file_path (str): abspath of the sqlite database file.

    Returns:
        dbout (sqlite connection): connection to the new sqlite database.
        dirname (str): abspath to the folder containing the sqlite database,
                       ie dir part of <file_path>.
        lib_name (str): library name, taken as the file name of <file_path>
                        without extension.
    """

    # make sure has .sqlite ext
    dirname,filename=os.path.split(file_path)
    lib_name,ext=os.path.splitext(filename)
    if ext=='':
        filename='%s.sqlite' %lib_name
        file_path=os.path.join(dirname,filename)

    dbfout=os.path.abspath(file_path)

    try:
        dbout = sqlite3.connect(dbfout)
        LOGGER.info('Connected to databaase %s' %file_path)
    except Exception:
        LOGGER.exception('Failed to connect to database %s' %file_path)

    cout=dbout.cursor()

    #dirname,filename=os.path.split(file_path)
    #lib_name,ext=os.path.splitext(filename)
    lib_folder=os.path.join(dirname,lib_name)

    if not os.path.exists(lib_folder):
        os.makedirs(lib_folder)
        LOGGER.info('Create lib folder %s' %lib_folder)

    file_folder=os.path.join(lib_folder,'_collections')
    if not os.path.exists(file_folder):
        os.makedirs(file_folder)
        LOGGER.info('Create lib collections folder %s' %file_folder)

    #--------------Create documents table--------------
    query='''CREATE TABLE IF NOT EXISTS Documents (
    id INTEGER PRIMARY KEY,
    %s)'''
    columns=[]
    for kii in DOC_ATTRS:
        if kii in INT_COLUMNS:
            columns.append('%s INT' %kii)
        else:
            columns.append('%s TEXT' %kii)

    columns=', '.join(columns)
    query=query %columns

    LOGGER.info('Creating empty table ...')
    cout.execute(query)
    #dbout.commit()

    #------------Create DocumentTags table------------
    query='''CREATE TABLE IF NOT EXISTS DocumentTags (
    did INT,
    tag TEXT)'''

    cout.execute(query)
    #dbout.commit()

    #------------Create DocumentNotes table------------
    query='''CREATE TABLE IF NOT EXISTS DocumentNotes (
    did INT,
    note TEXT,
    modifiedTime TEXT,
    createdTime TEXT
    )'''

    cout.execute(query)
    #dbout.commit()

    #----------Create DocumentKeywords table----------
    query='''CREATE TABLE IF NOT EXISTS DocumentKeywords (
    did INT,
    text TEXT)'''

    cout.execute(query)
    #dbout.commit()

    #-----------Create DocumentFolders table-----------
    query='''CREATE TABLE IF NOT EXISTS DocumentFolders (
    did INT,
    folderid INT
    )'''

    cout.execute(query)
    #dbout.commit()

    #---------------Create Folders table---------------
    query='''CREATE TABLE IF NOT EXISTS Folders (
    id INTEGER PRIMARY KEY,
    name TEXT,
    parentId INT,
    path TEXT,
    UNIQUE (name, parentId)
    )'''

    cout.execute(query)
    #dbout.commit()

    #--------Create DocumentContributors table--------
    query='''CREATE TABLE IF NOT EXISTS DocumentContributors (
    did INT,
    contribution TEXT,
    firstNames TEXT,
    lastName TEXT
    )'''

    cout.execute(query)
    #dbout.commit()

    #------------Create DocumentFiles table------------
    query='''CREATE TABLE IF NOT EXISTS DocumentFiles (
    did INT,
    relpath TEXT
    )'''

    cout.execute(query)
    #dbout.commit()

    #------------Create DocumentUrls table------------
    query='''CREATE TABLE IF NOT EXISTS DocumentUrls (
    did INT,
    url TEXT
    )'''

    cout.execute(query)
    #dbout.commit()

    #--------------create Default folder--------------
    query='''INSERT OR IGNORE INTO Folders (id, name, parentId, path)
    VALUES (?,?,?,?)'''

    #cout.execute(query, (0, 'Default', -1, os.path.join(lib_folder,'Default')))
    cout.execute(query, (0, 'Default', -1, os.path.join(lib_name,'Default')))
    dbout.commit()

    LOGGER.info('Created empty table.')

    #-----------------Add sample file-----------------
    #fname='./sample_bib.bib'
    #bib_entries=bibparse.readBibFile(fname)

    #print('# <createNewDatabase>: sample bib file.')
    #print(bib_entries)
    #print(type(bib_entries[0]))

    #doc=bib_entries[0]
    #doc['folders_l']=[(0,'Default')]
    #doc['files_l']=['./sample_pdf.pdf',]


    #metaDictToDatabase(dbout,1,bib_entries[0],lib_folder,rename_files)

    return dbout, dirname, lib_name


def metaDictToDatabase(db, docid, meta_dict, lib_folder, rename_files):
    """Save document changes to sqlite

    Args:
        db (sqlite connection): sqlite connection.
        docid (int): id of doc to save changes.
        meta_dict (DocMeta): meta data dict.
        lib_folder (str): abspath to the folder of the library. By design
                          this should point to the folder CONTAINING the
                          sqlite database file.
        rename_files (bool): whether to rename attachment files when saving.

    Returns: rec (int): 0 if success, None otherwise.

    3 types of changes are handled in this function:
        * insertion: <docid> is not found in sqlite, addToDatabase() is called.
        * deletion: <docid> in in sqlite, and <meta_dict> is None,
                    delDocFromDatabase() is called.
        * update: <docid> is in sqlite, and <meta_dict> is not None,
                  updateToDatabase() is called.
    """

    query='''SELECT DISTINCT rowid
    FROM Documents
    '''
    docids=db.execute(query).fetchall()
    docids=[ii[0] for ii in docids]

    if docid in docids:

        if meta_dict is None:
            LOGGER.info('docid %s in database. New meta=None. Deleting...' %docid)
            rec=delDocFromDatabase(db,docid,lib_folder)
        else:
            LOGGER.info('docid %s in database. Updating...' %docid)
            rec=updateToDatabase(db,docid,meta_dict,lib_folder,rename_files)
    else:
        if meta_dict is None:
            LOGGER.info('docid %s not in database. New meta=None. Ignore.' %docid)
            rec=0
        else:
            LOGGER.info('docid %s not in database. Inserting...' %docid)
            rec=addToDatabase(db,docid,meta_dict,lib_folder,rename_files)

    LOGGER.info('Done updating doc to database.')

    return rec


def insertToDocuments(db, docid, meta_dict, action):
    """Insert or update columns in the Documents table

    Args:
        db (sqlite connection): sqlite connection.
        docid (int): id of doc to save changes.
        meta_dict (DocMeta): meta data dict.
        action (str): if 'insert', use 'INSERT OR IGNORE' sqlite statement.
                      if 'replace', use 'REPLACE INTO' sqlite statement.

    Returns: 0
    """

    cout=db.cursor()
    #--------------Update Documents table--------------
    key_list=[]
    value_list=[]

    for kk,vv in meta_dict.items():

        if kk in DOC_ATTRS:
            LOGGER.debug('key = %s, value = %s' %(str(kk), str(vv)))
            key_list.append(kk)
            value_list.append(vv)

    if action=='insert':
        query='''INSERT OR IGNORE INTO Documents (rowid, %s) VALUES (%s)''' \
                %(', '.join(key_list), ','.join(['?',]*(1+len(key_list))))
    elif action=='replace':
        query='''REPLACE INTO Documents (rowid, %s) VALUES (%s)''' \
                %(', '.join(key_list), ','.join(['?',]*(1+len(key_list))))
    else:
        raise Exception("action not defined.")

    cout.execute(query,[docid,]+value_list)
    db.commit()

    LOGGER.info('Done inserting doc %s to Documents table.' %docid)

    return 0


def insertToTable(db, table, columns, values):
    """Insert columns in a given table

    Args:
        db (sqlite connection): sqlite connection.
        table (str): table name.
        columns (list): column names.
        values (list or tuple): new values to save to columns.

    Returns: 0
    """

    if len(values)>0:
        cout=db.cursor()
        query='INSERT OR IGNORE INTO %s (%s) VALUES (%s)' %\
                (table, ','.join(columns), ','.join(['?',]*len(columns)))
        for vii in values:
            cout.execute(query, vii)
        db.commit()

    LOGGER.info('Done inserting to %s table.' %table)

    return 0


def delFromTable(db, table, docid):
    """Delete table rows matching given doc id

    Args:
        db (sqlite connection): sqlite connection.
        table (str): table name.
        docid (int): doc id.

    Returns: 0
    """

    cout=db.cursor()
    query='DELETE FROM %s WHERE (%s.did=?)' %(table,table)
    cout.execute(query, (docid,))
    db.commit()
    LOGGER.debug('Deleted old rows in table %s with docid = %s' %(table,docid))

    return 0


def insertToDocumentFiles(db, docid, meta_dict, lib_folder, rename_files):
    """Insert or update columns in the DocumentsFiles table

    Args:
        db (sqlite connection): sqlite connection.
        docid (int): id of doc to save changes.
        meta_dict (DocMeta): meta data dict.
        lib_folder (str): abspath to the folder of the library. By design
                          this should point to the folder CONTAINING the
                          sqlite database file.
        rename_files (bool): whether to rename attachment files when saving.

    Returns: 0

    The path of each attachment file of the give doc is renamed if required,
    and copied to a specified folder corresponding to the library.
    A relative file path, relative to <lib_folder>, is obtained, and saved
    to sqlite.
    """

    cout=db.cursor()
    _,lib_name=os.path.split(lib_folder)
    if lib_name=='':
        lib_name=os.path.split(_)[1]

    abs_file_folder=os.path.join(lib_folder,'_collections')
    rel_file_folder=os.path.join('','_collections')
    LOGGER.debug('lib_name = %s' %lib_name)
    LOGGER.debug('abs_file_folder = %s' %abs_file_folder)
    LOGGER.debug('rel_file_folder = %s' %rel_file_folder)

    if not os.path.exists(abs_file_folder):
        os.makedirs(abs_file_folder)

    files=meta_dict['files_l']
    if len(files)>0:
        query='''INSERT OR IGNORE INTO DocumentFiles (did, relpath)
        VALUES (?,?)'''

        for fii in files:
            fii=os.path.expanduser(fii)
            fii=os.path.abspath(fii)

            if rename_files:
                filename=renameFile(fii,meta_dict)
            else:
                filename=os.path.split(fii)[1]

            #absii=tools.removeInvalidPathChar(absii)
            #folder,filename=os.path.split(absii)
            # can't do filename change here, will cause missing file.
            #filename=re.sub(r'[<>:"|?*]','_',filename)
            #if rename_files:
                #filename=renameFile(fii,meta_dict)
            #newabsii=os.path.join(lib_folder,'_collections')
            newabsii=os.path.join(abs_file_folder,filename)

            # deal with name conflicts
            newabsii=autoRename(newabsii)
            filename=os.path.split(newabsii)[1]

            rel_fii=os.path.join(rel_file_folder,filename)
            cout.execute(query,(docid,rel_fii))

            LOGGER.debug('new abspath = %s' %newabsii)
            LOGGER.debug('new relpath = %s' %rel_fii)

            #if not os.path.exists(newabsii):
            # check if fii is abspath or relpath
            if os.path.isabs(fii):
                try:
                    shutil.copy(fii,newabsii)
                    LOGGER.info('copy file %s -> %s' %(fii,newabsii))
                except:
                    LOGGER.exception('Failed to copy file %s to %s' %(fii,newabsii))

            #--------------------Copy file--------------------
            #try:
                #shutil.copy(fii, newabsii)
                #LOGGER.debug('Copied file %s to %s' %(absii,newabsii))
            #except:
                #LOGGER.exception('Failed to copy file %s to %s' %(absii,newabsii))

        db.commit()

    LOGGER.info('Done inserting doc %s to DocumentFiles' %docid)

    return 0


def addToDatabase(db, docid, meta_dict, lib_folder, rename_files):
    """Add new document to sqlite

    Args:
        db (sqlite connection): sqlite connection.
        docid (int): id of doc to save changes.
        meta_dict (DocMeta): meta data dict.
        lib_folder (str): abspath to the folder of the library. By design
                          this should point to the folder CONTAINING the
                          sqlite database file.
        rename_files (bool): whether to rename attachment files when saving.

    Returns: rec (int): 0 if success, None otherwise.
    """

    LOGGER.info('Adding doc to database. docid = %s' %docid)

    #--------------Update Documents table--------------
    rec=insertToDocuments(db, docid, meta_dict, 'insert')
    LOGGER.debug('rec of insertToDocuments = %s' %rec)

    #------------------Update authors------------------
    firsts=meta_dict['firstNames_l']

    if len(firsts)>0:
        lasts=meta_dict['lastName_l']
        rec=insertToTable(db, 'DocumentContributors',
                ('did', 'contribution', 'firstNames', 'lastName'),
                list(zip((docid,)*len(firsts),
                    ('DocumentAuthor',)*len(firsts),
                    firsts, lasts))
                )
        LOGGER.debug('rec of insertToDocumentAuthors = %s' %rec)

    #-------------------Update files-------------------
    rec=insertToDocumentFiles(db, docid, meta_dict, lib_folder, rename_files)
    LOGGER.debug('rec of insertToDocumentFiles = %s' %rec)

    #------------------Update folder------------------
    rec=insertToTable(db, 'DocumentFolders',
            ('did', 'folderid'),
            [(docid,fii[0]) for fii in meta_dict['folders_l']])
    LOGGER.debug('rec of insertDocumentFolders = %s' %rec)

    #-----------------Update keywords-----------------
    rec=insertToTable(db, 'DocumentKeywords',
            ('did', 'text'),
            [(docid, kii) for kii in meta_dict['keywords_l']])
    LOGGER.debug('rec of insertDocumentKeywords = %s' %rec)

    #-------------------Update notes-------------------
    notes=meta_dict['notes']
    if notes:
        ctime=datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        rec=insertToTable(db, 'DocumentNotes',
                ('did', 'note', 'modifiedTime', 'createdTime'),
                [(docid, notes, ctime, ctime)]
                )
    LOGGER.debug('rec of insertDocumentNotes = %s' %rec)

    #-----------------Update tags-----------------
    rec=insertToTable(db, 'DocumentTags',
            ('did', 'tag'),
            [(docid, kii) for kii in meta_dict['tags_l']])
    LOGGER.debug('rec of insertDocumentTags = %s' %rec)

    #-------------------Update urls-------------------
    rec=insertToTable(db, 'DocumentUrls',
            ('did', 'url'),
            [(docid, kii) for kii in meta_dict['urls_l']])
    LOGGER.debug('rec of insertDocumentUrls = %s' %rec)

    LOGGER.info('Done adding doc to database.')

    return 0


def updateToDatabase(db, docid, meta_dict, lib_folder, rename_files):
    """Save changes of existing document to sqlite

    Args:
        db (sqlite connection): sqlite connection.
        docid (int): id of doc to save changes.
        meta_dict (DocMeta): meta data dict.
        lib_folder (str): abspath to the folder of the library. By design
                          this should point to the folder CONTAINING the
                          sqlite database file.
        rename_files (bool): whether to rename attachment files when saving.

    Returns: rec (int): 0 if success, None otherwise.
    """

    cout=db.cursor()
    LOGGER.info('Updating doc to database. docid = %s' %docid)

    #----------Get a meta_dict from database----------
    old_meta=getMetaData(db, docid)

    #------------------Update folders------------------
    folders=meta_dict['folders_l']
    old_folders=old_meta['folders_l']
    LOGGER.debug('previous folders = %s. current folders = %s'\
            %(old_folders, folders))

    if set(old_folders) != set(folders):

        del_folders=list(set(old_folders).difference(folders))
        new_folders=list(set(folders).difference(old_folders))
        LOGGER.debug('Need to update folders.')
        LOGGER.debug('del folders = %s. new folders = %s'\
                %(del_folders, new_folders))

        for fii in del_folders:

            query='''DELETE FROM DocumentFolders
            WHERE (DocumentFolders.did=? AND DocumentFolders.folderid=?)
            '''
            cout.execute(query, (docid, int(fii[0])))

        for fii in new_folders:

            query='''INSERT OR IGNORE INTO DocumentFolders (did, folderid)
            VALUES (?,?)'''
            cout.execute(query, (docid, int(fii[0])))

    #--------------Update Documents table--------------
    rec=insertToDocuments(db, docid, meta_dict, 'replace')
    LOGGER.debug('rec of insertToDocuments = %s' %rec)

    #------------------Update authors------------------
    if old_meta['authors_l']!=meta_dict['authors_l']:
        # order change is also a change
        LOGGER.debug('Need to update authors.')

        #----------------Remove old authors----------------
        delFromTable(db, 'DocumentContributors', docid)

        #------------------Update authors------------------
        firsts=meta_dict['firstNames_l']

        if len(firsts)>0:
            lasts=meta_dict['lastName_l']
            rec=insertToTable(db, 'DocumentContributors',
                    ('did', 'contribution', 'firstNames', 'lastName'),
                    list(zip((docid,)*len(firsts),
                        ('DocumentAuthor',)*len(firsts),
                        firsts, lasts))
                    )
            LOGGER.debug('rec of insertToDocumentAuthors = %s' %rec)

    #-------------------Update files-------------------
    if set(old_meta['files_l']) != set(meta_dict['files_l']):

        LOGGER.debug('Need to update files.')
        delFromTable(db, 'DocumentFiles', docid)
        rec=insertToDocumentFiles(db, docid, meta_dict, lib_folder, rename_files)
        LOGGER.debug('rec of insertToDocumentFiles=%s' %rec)

    #-----------------Update keywords-----------------
    if set(old_meta['keywords_l']) != set(meta_dict['keywords_l']):

        LOGGER.debug('Need to update keywords.')
        delFromTable(db, 'DocumentKeywords', docid)
        rec=insertToTable(db, 'DocumentKeywords',
                ('did', 'text'),
                [(docid, kii) for kii in meta_dict['keywords_l']])
        LOGGER.debug('rec of insertDocumentKeywords = %s' %rec)

    #-------------------Update notes-------------------
    notes=meta_dict['notes']
    if old_meta['notes'] != notes:

        LOGGER.debug('Need to update notes.')

        if notes:
            mtime=datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
            # Get createdTime if old note exists
            query='''
            SELECT (DocumentNotes.createdTime) FROM DocumentNotes
            WHERE (DocumentNotes.did=?)
            '''
            ctime=fetchField(db,query,(docid,),1,'str')

            if ctime is None:
                LOGGER.debug('Old creation time not exists. Use current time.')
                ctime=mtime

            delFromTable(db, 'DocumentNotes', docid)

            rec=insertToTable(db, 'DocumentNotes',
                    ('did', 'note', 'modifiedTime', 'createdTime'),
                    [(docid, notes, mtime, ctime)]
                    )
            LOGGER.debug('rec of insertDocumentNotes = %s' %rec)
        else:
            delFromTable(db, 'DocumentNotes', docid)

    #-------------------Update tags-------------------
    if set(old_meta['tags_l']) != set(meta_dict['tags_l']):

        LOGGER.debug('Need to update tags.')

        delFromTable(db, 'DocumentTags', docid)
        rec=insertToTable(db, 'DocumentTags',
                ('did', 'tag'),
                [(docid, kii) for kii in meta_dict['tags_l']])
        LOGGER.debug('rec of insertDocumentTags = %s' %rec)

    #-------------------Update urls-------------------
    if set(old_meta['urls_l']) != set(meta_dict['urls_l']):

        LOGGER.debug('Need to update urls.')

        delFromTable(db, 'DocumentUrls', docid)
        rec=insertToTable(db, 'DocumentUrls',
                ('did', 'url'),
                [(docid, kii) for kii in meta_dict['urls_l']])
        LOGGER.debug('rec of insertDocumentUrls = %s' %rec)

    LOGGER.info('Done updating doc to database.')

    return 0


def delDocFromDatabase(db, docid, lib_folder):
    """Delete existing document from sqlite

    Args:
        db (sqlite connection): sqlite connection.
        docid (int): id of doc to delete.
        lib_folder (str): abspath to the folder of the library. By design
                          this should point to the folder CONTAINING the
                          sqlite database file.

    Returns: rec (int): 0 if success, None otherwise.
    """

    cout=db.cursor()
    LOGGER.info('Deleting doc %s from database' %docid)

    #--------------del Documents table--------------
    query='DELETE FROM Documents WHERE (Documents.id=?)'
    cout.execute(query, (docid,))
    #db.commit()

    #------------------del folders------------------
    delFromTable(db, 'DocumentFolders', docid)

    #------------------del authors------------------
    delFromTable(db, 'DocumentContributors', docid)

    #-------------------del files-------------------
    query_files='''
    SELECT DocumentFiles.relpath
    FROM DocumentFiles
    WHERE (DocumentFiles.did=?)
    '''
    old_files=fetchField(db, query_files, (docid,), 1, 'list')
    delFromTable(db, 'DocumentFiles', docid)

    for fii in old_files:
        # prepend folder path
        absii=os.path.join(lib_folder,fii)
        if os.path.exists(absii):
            LOGGER.info('Deleting file from disk %s' %absii)
            os.remove(absii)

    #-----------------del keywords-----------------
    delFromTable(db, 'DocumentKeywords', docid)

    #-------------------del notes-------------------
    delFromTable(db, 'DocumentNotes', docid)

    #-------------------del tags-------------------
    delFromTable(db, 'DocumentTags', docid)

    #-------------------del urls-------------------
    delFromTable(db, 'DocumentUrls', docid)

    LOGGER.info('Done deleting doc from database.')

    return 0
